from typing import List, Union
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.modelMonitor import ModelMonitor, ModelMonitorCreateDto
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.schemas.user import User
from whitebox.utils.errors import add_error_responses, errors


model_monitors_router = APIRouter()


@model_monitors_router.post(
    "/model-monitors",
    tags=["Model Monitors"],
    response_model=ModelMonitor,
    summary="Create a model monitor",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 401, 409]),
)
async def create_model_monitor(
    body: ModelMonitorCreateDto,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> ModelMonitor:
    """Inserts a model monitor into the database."""

    new_model_monitor = crud.model_monitors.create(db=db, obj_in=body)
    return new_model_monitor


@model_monitors_router.get(
    "/model-monitors",
    tags=["Model Monitors"],
    response_model=List[ModelMonitor],
    summary="Get all model's model monitors",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_all_models_model_monitors(
    model_id: Union[str, None] = None,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """
    Fetches model monitors from the databse.
    \n If a model id is provided, only the monitors for the specific model will be fetched.
    \n If a model id is not provided then all monitors from the database will be fetched.
    """

    if model_id:
        model = crud.models.get(db, model_id)
        if model:
            return crud.model_monitors.get_model_monitors_by_model(
                db=db, model_id=model_id
            )
        else:
            return errors.not_found("Model not found")
    else:
        return crud.model_monitors.get_all(db=db)



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from whitebox.api.v1 import v1_router
from fastapi.openapi.utils import get_openapi
from whitebox.api.v1.docs import (
    tags_metadata,
    validation_error,
    authorization_error,
    not_found_error,
    conflict_error,
    content_gone,
    bad_request,
)
from whitebox.core.settings import get_settings
from whitebox.core.db import connect, close
from starlette.exceptions import HTTPException as StarletteHTTPException
from fastapi.exceptions import RequestValidationError

from whitebox.utils.errors import errors

settings = get_settings()

app = FastAPI(title=settings.APP_NAME, redoc_url="/")

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(v1_router)


app.add_exception_handler(StarletteHTTPException, errors.http_exception_handler)
app.add_exception_handler(RequestValidationError, errors.validation_exception_handler)


@app.on_event("startup")
async def on_app_start():
    """Anything that needs to be done while app starts"""
    await connect()


@app.on_event("shutdown")
async def on_app_shutdown():
    """Anything that needs to be done while app shutdown"""
    await close()


def app_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="Whitebox",
        version=settings.VERSION,
        routes=app.routes,
        tags=tags_metadata,
    )

    openapi_schema["components"]["schemas"]["HTTPValidationError"] = validation_error
    openapi_schema["components"]["schemas"]["AuthorizationError"] = authorization_error
    openapi_schema["components"]["schemas"]["NotFoundError"] = not_found_error
    openapi_schema["components"]["schemas"]["ConflictError"] = conflict_error
    openapi_schema["components"]["schemas"]["BadRequest"] = bad_request
    openapi_schema["components"]["schemas"]["ContentGone"] = content_gone

    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = app_openapi



import pandas as pd
from typing import List
from whitebox.analytics.models.pipelines import (
    create_binary_classification_training_model_pipeline,
    create_multiclass_classification_training_model_pipeline,
    create_regression_training_model_pipeline,
)
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.datasetRow import DatasetRow, DatasetRowCreate
from whitebox.schemas.model import ModelType
from whitebox.schemas.user import User
from fastapi import APIRouter, BackgroundTasks, Depends, status
from fastapi.encoders import jsonable_encoder
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.utils.errors import add_error_responses, errors


dataset_rows_router = APIRouter()


@dataset_rows_router.post(
    "/dataset-rows",
    tags=["Dataset Rows"],
    response_model=List[DatasetRow],
    summary="Create dataset rows",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 401, 404, 409]),
)
async def create_dataset_rows(
    body: List[DatasetRowCreate],
    background_tasks: BackgroundTasks,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> DatasetRow:
    """
    Inserts a set of dataset rows into the database.
    \nWhen the dataset rows are successfully saved, the pipeline for training the model is triggered.
    """

    if len(body) <= 1:
        return errors.bad_request("Training dataset should be longer that 1 row!")

    model = crud.models.get(db=db, _id=dict(body[0])["model_id"])
    if model:
        for row in body:
            if not model.prediction in row.processed:
                return errors.bad_request(
                    f'Column "{model.prediction}" was not found in some or any of the rows in provided training dataset. Please try again!'
                )

        predictions = list(set(vars(x)["processed"][model.prediction] for x in body))
        if len(predictions) <= 1:
            return errors.bad_request(
                f'Training dataset\'s "{model.prediction}" columns must have at least 2 different values!'
            )

        new_dataset_rows = crud.dataset_rows.create_many(db=db, obj_list=body)
        processed_dataset_rows = [
            x["processed"] for x in jsonable_encoder(new_dataset_rows)
        ]
        processed_dataset_rows_pd = pd.DataFrame(processed_dataset_rows)

        if model.type == ModelType.binary:
            background_tasks.add_task(
                create_binary_classification_training_model_pipeline,
                processed_dataset_rows_pd,
                model.prediction,
                model.id,
            )
        elif model.type == ModelType.multi_class:
            background_tasks.add_task(
                create_multiclass_classification_training_model_pipeline,
                processed_dataset_rows_pd,
                model.prediction,
                model.id,
            )
        elif model.type == ModelType.regression:
            background_tasks.add_task(
                create_regression_training_model_pipeline,
                processed_dataset_rows_pd,
                model.prediction,
                model.id,
            )
        return new_dataset_rows
    else:
        return errors.not_found(f"Model with id: {dict(body[0])['model_id']} not found")


@dataset_rows_router.get(
    "/dataset-rows",
    tags=["Dataset Rows"],
    response_model=List[DatasetRow],
    summary="Get all model's dataset rows",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_all_dataset_rows(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches the dataset rows of a specific model. A model id is required."""

    model = crud.models.get(db, model_id)
    if model:
        return crud.dataset_rows.get_dataset_rows_by_model(db=db, model_id=model_id)
    else:
        return errors.not_found("Model not found")



from .drifting_metrics import *
from .dataset_rows import *
from .inference_rows import *
from .model_integrity_metrics import *
from .model_monitors import *
from .performance_metrics import *
from .alerts import *
from .users import *
from .models import *
from .base import *



from typing import Any, List
from sqlalchemy.orm import Session
from sqlalchemy import desc
from whitebox.crud.base import CRUDBase
from whitebox.entities.DriftingMetric import DriftingMetric as DriftingMetricEntity
from whitebox.schemas.driftingMetric import DriftingMetric


class CRUD(CRUDBase[DriftingMetric, Any, Any]):
    def get_drifting_metrics_by_model(
        self, db: Session, *, model_id: str
    ) -> List[DriftingMetric]:
        return (
            db.query(self.model).filter(DriftingMetricEntity.model_id == model_id).all()
        )

    def get_latest_report_by_model(
        self, db: Session, *, model_id: int
    ) -> DriftingMetric:
        return (
            db.query(self.model)
            .filter(self.model.model_id == model_id)
            .order_by(desc("created_at"))
            .first()
        )


drifting_metrics = CRUD(DriftingMetricEntity)



from typing import Any, List
from sqlalchemy.orm import Session
from whitebox.crud.base import CRUDBase
from whitebox.entities.Alert import Alert as AlertEntity
from whitebox.schemas.alert import Alert


class CRUD(CRUDBase[Alert, Any, Any]):
    def get_model_alerts_by_model(self, db: Session, *, model_id: str) -> List[Alert]:
        return db.query(self.model).filter(AlertEntity.model_id == model_id).all()


alerts = CRUD(AlertEntity)



from typing import List
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.modelIntegrityMetric import ModelIntegrityMetric
from whitebox.schemas.user import User
from whitebox.utils.errors import add_error_responses, errors


model_integrity_metrics_router = APIRouter()


@model_integrity_metrics_router.get(
    "/model-integrity-metrics",
    tags=["Model Integrity Metrics"],
    response_model=List[ModelIntegrityMetric],
    summary="Get all model's model integrity metrics",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_all_models_model_integrity_metrics(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches the model integrity metrics of a specific model. A model id is required."""

    model = crud.models.get(db, model_id)
    if model:
        return crud.model_integrity_metrics.get_model_integrity_metrics_by_model(
            db=db, model_id=model_id
        )
    else:
        return errors.not_found("Model not found")



from sqlalchemy import Column, String, ForeignKey, DateTime
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid


class Alert(Base):
    __tablename__ = "alerts"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    model_monitor_id = Column(
        String, ForeignKey("model_monitors.id", ondelete="CASCADE")
    )
    timestamp = Column(DateTime)
    description = Column(String)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)



from .inferenceRow import *
from .model import *
from .alert import *
from .base import *
from .datasetRow import *
from .driftingMetric import *
from .modelIntegrityMetric import *
from .modelMonitor import *
from .performanceMetric import *
from .user import *
from .utils import *



from typing import Any, List, Union
from sqlalchemy.orm import Session
from sqlalchemy import desc
from whitebox.crud.base import CRUDBase
from whitebox.entities.PerformanceMetric import (
    BinaryClassificationMetrics as BinaryClassificationMetricsEntity,
    MultiClassificationMetrics as MultiClassificationMetricsEntity,
    RegressionMetrics as RegressionMetricsEntity,
)
from whitebox.schemas.performanceMetric import (
    BinaryClassificationMetrics,
    MultiClassificationMetrics,
    RegressionMetrics,
)


class CRUD(
    CRUDBase[Union[BinaryClassificationMetrics, MultiClassificationMetrics], Any, Any]
):
    def get_by_model(
        self, db: Session, *, model_id: int
    ) -> Union[
        List[BinaryClassificationMetrics],
        List[MultiClassificationMetrics],
        List[RegressionMetrics],
    ]:
        return db.query(self.model).filter(self.model.model_id == model_id).all()

    def get_latest_report_by_model(
        self, db: Session, *, model_id: int
    ) -> Union[
        BinaryClassificationMetrics, MultiClassificationMetrics, RegressionMetrics
    ]:
        return (
            db.query(self.model)
            .filter(self.model.model_id == model_id)
            .order_by(desc("created_at"))
            .first()
        )


binary_classification_metrics = CRUD(BinaryClassificationMetricsEntity)
multi_classification_metrics = CRUD(MultiClassificationMetricsEntity)
regression_metrics = CRUD(RegressionMetricsEntity)



from pydantic import BaseModel


class HealthCheck(BaseModel):
    status: str


class StatusCode(BaseModel):
    status_code: str


class ErrorProps(BaseModel):
    error: str
    status_code: int


class ErrorResponse(BaseModel):
    title: str
    type: str
    properties: ErrorProps



from functools import lru_cache
from pydantic import BaseSettings
import os


class Settings(BaseSettings):
    APP_NAME: str = ""
    ENV: str = ""
    DATABASE_URL: str = ""
    VERSION: str = ""
    MODEL_PATH: str = ""
    SECRET_KEY: str = ""

    class Config:
        env_file = f".env.{os.getenv('ENV')}" or ".env.dev"


@lru_cache()
def get_settings():
    return Settings()


class CronSettings(Settings):
    APP_NAME_CRON: str
    METRICS_CRON: str

    class Config:
        env_file = f".env.{os.getenv('ENV')}" or ".env.dev"


@lru_cache()
def get_cron_settings():
    return CronSettings()



from asyncio.tasks import Task
from dataclasses import dataclass

from pydantic import BaseModel, Field
from typing import Callable, Coroutine, Optional, List, Dict, Deque, Literal, Union
from uuid import UUID, uuid4

import datetime
import pytz


def now():
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


TaskStatus = Literal[
    "registered", "running", "finished", "pending", "cancelled", "failed"
]

EventType = Literal[
    "task_registered",
    "task_started",
    "task_failed",
    "task_finished",
    "task_cancelled",
    "task_disabled",
]


class TaskInfo(BaseModel):
    uid: UUID = uuid4()
    name: str
    status: str
    previous_status: str
    enabled: bool
    crontab: Union[str, None] = None
    created_at: Union[datetime.datetime, float] = now().timestamp()
    started_at: Union[datetime.datetime, float, None] = None
    stopped_at: Union[datetime.datetime, float, None] = None
    next_run_in: Union[int, None] = None


class TaskDefinition(BaseModel):
    name: str
    async_callable: Callable[[], Coroutine]
    enabled: bool = True
    crontab: Optional[str] = None


@dataclass
class RunningTask:
    task_definition: TaskDefinition
    asyncio_task: Task
    since: Union[datetime.datetime, float]


class TaskLog(BaseModel):
    event_type: EventType
    task_name: str
    crontab: Union[str, None] = None
    enabled: bool
    error: Union[str, None] = None
    timestamp: int = Field(default_factory=lambda: datetime.datetime.now().timestamp())


class State(BaseModel):
    created_at: datetime.datetime
    tasks_info: List[TaskInfo]


class TaskRealTimeInfo(BaseModel):
    name: str
    status: TaskStatus
    previous_status: Optional[TaskStatus]
    next_run_ts: Optional[int]
    started_at: Optional[Union[datetime.datetime, float]]
    stopped_at: Optional[Union[datetime.datetime, float]]



from fastapi import APIRouter
from .health import health_router

from .models import models_router
from .dataset_rows import dataset_rows_router
from .inference_rows import inference_rows_router
from .performance_metrics import performance_metrics_router
from .drifting_metrics import drifting_metrics_router
from .model_integrity_metrics import model_integrity_metrics_router
from .model_monitors import model_monitors_router
from .alerts import alerts_router
from .cron_tasks import cron_tasks_router


v1_router = APIRouter()
v1 = "/v1"

v1_router.include_router(health_router, prefix=v1)
v1_router.include_router(models_router, prefix=v1)
v1_router.include_router(dataset_rows_router, prefix=v1)
v1_router.include_router(inference_rows_router, prefix=v1)
v1_router.include_router(performance_metrics_router, prefix=v1)
v1_router.include_router(drifting_metrics_router, prefix=v1)
v1_router.include_router(model_integrity_metrics_router, prefix=v1)
v1_router.include_router(model_monitors_router, prefix=v1)
v1_router.include_router(alerts_router, prefix=v1)
v1_router.include_router(cron_tasks_router, prefix=v1)



from whitebox.tests.v1.mock_data import (
    dataset_rows_single_row_column_payload,
    dataset_rows_no_prediction_column_payload,
    dataset_rows_one_prediction_value_payload,
    dataset_rows_create_multi_class_payload,
    dataset_rows_create_binary_payload,
    dataset_rows_create_wrong_model_payload,
    dataset_rows_create_reg_payload,
)
import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("dataset_rows_wrong_training_dataset"))
def test_dataset_rows_wrong_training_data(client, api_key):
    response_single_row = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi["id"]},
                dataset_rows_single_row_column_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_no_prediction = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi["id"]},
                dataset_rows_no_prediction_column_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_one_prediction_value = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi["id"]},
                dataset_rows_one_prediction_value_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    assert response_single_row.status_code == status.HTTP_400_BAD_REQUEST
    assert response_no_prediction.status_code == status.HTTP_400_BAD_REQUEST
    assert response_one_prediction_value.status_code == status.HTTP_400_BAD_REQUEST


@pytest.mark.order(get_order_number("dataset_rows_create"))
def test_dataset_row_create_many(client, api_key):
    response_model_multi = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi["id"]},
                dataset_rows_create_multi_class_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_model_binary = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_binary["id"]},
                dataset_rows_create_binary_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_model_reg = client.post(
        "/v1/dataset-rows",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_regression["id"]},
                dataset_rows_create_reg_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_wrong_model = client.post(
        "/v1/dataset-rows",
        json=(dataset_rows_create_wrong_model_payload),
        headers={"api-key": api_key},
    )

    assert response_model_multi.status_code == status.HTTP_201_CREATED
    assert response_model_binary.status_code == status.HTTP_201_CREATED
    assert response_model_reg.status_code == status.HTTP_201_CREATED
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND
    validated = [schemas.DatasetRow(**m) for m in response_model_multi.json()]
    validated = [schemas.DatasetRow(**m) for m in response_model_binary.json()]
    validated = [schemas.DatasetRow(**m) for m in response_model_reg.json()]


@pytest.mark.order(get_order_number("dataset_rows_get_model's_all"))
def test_dataset_row_get_models_all(client, api_key):
    response_model_multi = client.get(
        f"/v1/dataset-rows?model_id={state.model_multi['id']}",
        headers={"api-key": api_key},
    )
    response_model_binary = client.get(
        f"/v1/dataset-rows?model_id={state.model_binary['id']}",
        headers={"api-key": api_key},
    )
    response_model_not_found = client.get(
        f"/v1/dataset-rows?model_id=wrong_model_id",
        headers={"api-key": api_key},
    )

    assert response_model_multi.status_code == status.HTTP_200_OK
    assert response_model_binary.status_code == status.HTTP_200_OK
    assert response_model_not_found.status_code == status.HTTP_404_NOT_FOUND
    validated = [schemas.DatasetRow(**m) for m in response_model_multi.json()]
    validated = [schemas.DatasetRow(**m) for m in response_model_binary.json()]



from datetime import datetime
from pydantic import BaseModel
from typing import Dict, Union
from whitebox.schemas.base import ItemBase

# TODO: Include comments of what each class represents


class BinaryClassificationMetricsPipelineResult(BaseModel):
    """This class is used to store the results of the pipeline that calculates the binary classification metrics"""

    accuracy: float
    precision: float
    recall: float
    f1: float
    true_negative: int
    false_positive: int
    false_negative: int
    true_positive: int


class BinaryClassificationMetricsBase(BinaryClassificationMetricsPipelineResult):
    model_id: str
    timestamp: Union[str, datetime]


class BinaryClassificationMetrics(BinaryClassificationMetricsBase, ItemBase):
    pass


class DifferentStatistics(BaseModel):
    micro: float
    macro: float
    weighted: float


class ConfusionMatrix(BaseModel):
    true_negative: int
    false_positive: int
    false_negative: int
    true_positive: int


class MultiClassificationMetricsPipelineResult(BaseModel):
    """This class is used to store the results of the pipeline that calculates the multi classification metrics"""

    accuracy: float
    precision: DifferentStatistics
    recall: DifferentStatistics
    f1: DifferentStatistics
    confusion_matrix: Dict[str, ConfusionMatrix]


class MultiClassificationMetricsBase(MultiClassificationMetricsPipelineResult):
    model_id: str
    timestamp: Union[str, datetime]


class MultiClassificationMetrics(MultiClassificationMetricsBase, ItemBase):
    pass


class RegressionMetricsPipelineResult(BaseModel):
    """This class is used to store the results of the pipeline that calculates the regression metrics"""

    r_square: float
    mean_squared_error: float
    mean_absolute_error: float


class RegressionMetricsBase(RegressionMetricsPipelineResult):
    model_id: str
    timestamp: Union[str, datetime]


class RegressionMetrics(RegressionMetricsBase, ItemBase):
    pass



from pydantic import BaseModel
from whitebox.schemas.base import ItemBase
from typing import Dict, Any


class DatasetRowBase(BaseModel):
    model_id: str
    # Data before any processing
    nonprocessed: Dict[str, Any]
    # Before model entry
    processed: Dict[str, float]


class DatasetRow(DatasetRowBase, ItemBase):
    pass


class DatasetRowCreate(DatasetRowBase):
    pass



from typing import Optional
from pydantic import BaseModel
import enum
from whitebox.schemas.base import ItemBase


class MonitorStatus(str, enum.Enum):
    active = "active"
    inactive = "inactive"


class MonitorMetrics(str, enum.Enum):
    # Performance metrics
    accuracy = "accuracy"
    precision = "precision"
    recall = "recall"
    f1 = "f1"
    r_square = "r_square"
    mean_squared_error = "mean_squared_error"
    mean_absolute_error = "mean_absolute_error"

    # Drifting metrics
    data_drift = "data_drift"
    concept_drift = "concept_drift"


class AlertSeverity(str, enum.Enum):
    low = "low"
    mid = "mid"
    high = "high"


class ModelMonitorBase(BaseModel):
    model_id: str
    name: str
    status: MonitorStatus
    metric: MonitorMetrics
    severity: AlertSeverity
    email: str
    feature: Optional[str]
    lower_threshold: Optional[float]


class ModelMonitor(ModelMonitorBase, ItemBase):
    pass


class ModelMonitorCreateDto(ModelMonitorBase):
    pass



from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.backends import default_backend
import secrets
from whitebox.core.settings import get_settings


settings = get_settings()


def to_utf8(ps):
    return ps.encode("utf-8")


def encrypt_api_key(password: str, key: bytes) -> str:
    iv = secrets.token_bytes(16)
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
    encryptor = cipher.encryptor()
    padder = padding.PKCS7(128).padder()
    padded_data = padder.update(to_utf8(password)) + padder.finalize()
    ct = encryptor.update(padded_data) + encryptor.finalize()
    return (iv + ct).hex()


def decrypt_api_key(password: str, key: bytes) -> str:
    ct = bytes.fromhex(password)
    iv = ct[:16]
    ct = ct[16:]
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
    decryptor = cipher.decryptor()
    pt = decryptor.update(ct) + decryptor.finalize()
    unpadder = padding.PKCS7(128).unpadder()
    data = unpadder.update(pt) + unpadder.finalize()
    return data.decode("utf-8")


def passwords_match(api_key_in_db: str, api_key_in_headers: str):
    if not settings.SECRET_KEY:
        return api_key_in_db == api_key_in_headers

    api_key = decrypt_api_key(api_key_in_db, settings.SECRET_KEY.encode())
    return api_key == api_key_in_headers



import pandas as pd
import pytest
from whitebox.schemas.modelMonitor import AlertSeverity, MonitorMetrics, MonitorStatus
from whitebox.sdk import Whitebox
from whitebox.tests.v1.conftest import get_order_number, state, state_sdk
from whitebox.tests.v1.mock_data import (
    model_multi_create_payload,
    timestamps,
    mixed_actuals,
    inference_row_xai_payload,
    drifting_metrics_report_payload,
    descriptive_statistics_report_payload,
    performance_metrics_report_payload,
)
import requests_mock
from fastapi import status


@pytest.mark.order(get_order_number("sdk_init"))
def test_sdk_init(client, api_key):
    wb = Whitebox(host=client.base_url, api_key=api_key)
    assert wb.host == client.base_url
    assert wb.api_key == api_key
    state_sdk.wb = wb


@pytest.mark.order(get_order_number("sdk_create_model"))
def test_sdk_create_model(client):
    with requests_mock.Mocker() as m:
        m.post(
            url=f"{state_sdk.wb.host}/v1/models",
            json=model_multi_create_payload,
            headers={"api-key": state_sdk.wb.api_key},
        )

        model = state_sdk.wb.create_model(
            name=model_multi_create_payload["name"],
            description=model_multi_create_payload["description"],
            labels=model_multi_create_payload["labels"],
            type=model_multi_create_payload["type"],
            prediction=model_multi_create_payload["prediction"],
        )

        assert model == model_multi_create_payload


@pytest.mark.order(get_order_number("sdk_get_model"))
def test_sdk_get_model(client):
    mock_model_id = "mock_model_id"
    with requests_mock.Mocker() as m:
        m.get(
            url=f"{state_sdk.wb.host}/v1/models/{mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_404_NOT_FOUND,
        )

        not_found_result = state_sdk.wb.get_model(model_id=mock_model_id)

        assert not_found_result == None

        m.get(
            url=f"{state_sdk.wb.host}/v1/models/{mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            json=state.model_multi,
        )

        model = state_sdk.wb.get_model(model_id=mock_model_id)

        assert model == state.model_multi


@pytest.mark.order(get_order_number("sdk_delete_model"))
def test_sdk_delete_model(client):
    mock_model_id = "mock_model_id"

    with requests_mock.Mocker() as m:
        m.delete(
            url=f"{state_sdk.wb.host}/v1/models/{mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_200_OK,
        )

        happy_result = state_sdk.wb.delete_model(model_id=mock_model_id)
        assert happy_result == True

        m.delete(
            url=f"{state_sdk.wb.host}/v1/models/{mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        sad_result = state_sdk.wb.delete_model(model_id=mock_model_id)
        assert sad_result == False


@pytest.mark.order(get_order_number("sdk_log_training_dataset"))
def test_sdk_log_training_dataset(client):
    mock_model_id = "mock_model_id"
    df = pd.read_csv("whitebox/analytics/data/testing/classification_test_data.csv")

    with requests_mock.Mocker() as m:
        m.post(
            url=f"{state_sdk.wb.host}/v1/dataset-rows",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_201_CREATED,
        )

        happy_result = state_sdk.wb.log_training_dataset(
            model_id=mock_model_id, processed=df, non_processed=df
        )
        assert happy_result == True

        m.post(
            url=f"{state_sdk.wb.host}/v1/dataset-rows",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        sad_result = state_sdk.wb.log_training_dataset(
            model_id=mock_model_id, processed=df, non_processed=df
        )
        assert sad_result == False

    # drop a row in df to test the dataframe length error handling
    df2 = df.drop(df.index[0])
    with pytest.raises(Exception) as e_info:
        state_sdk.wb.log_inferences(
            model_id=mock_model_id, processed=df, non_processed=df2
        )


@pytest.mark.order(get_order_number("sdk_log_inferences"))
def test_sdk_log_inferences(client):
    mock_model_id = "mock_model_id"
    df = pd.read_csv("whitebox/analytics/data/testing/classification_test_data.csv")

    with requests_mock.Mocker() as m:
        m.post(
            url=f"{state_sdk.wb.host}/v1/inference-rows/batch",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_201_CREATED,
        )

        happy_result = state_sdk.wb.log_inferences(
            model_id=mock_model_id,
            processed=df,
            non_processed=df,
            timestamps=timestamps,
            actuals=mixed_actuals,
        )
        assert happy_result == True

        m.post(
            url=f"{state_sdk.wb.host}/v1/inference-rows/batch",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        sad_result = state_sdk.wb.log_inferences(
            model_id=mock_model_id,
            processed=df,
            non_processed=df,
            timestamps=timestamps,
        )
        assert sad_result == False

    # drop a row in df to test the dataframe length error handling
    df2 = df.drop(df.index[0])
    with pytest.raises(Exception) as e_info:
        state_sdk.wb.log_inferences(
            model_id=mock_model_id,
            processed=df,
            non_processed=df2,
            timestamps=timestamps,
            actuals=mixed_actuals,
        )


@pytest.mark.order(get_order_number("sdk_create_model_monitor"))
def test_sdk_create_model_monitor(client):
    with requests_mock.Mocker() as m:
        m.post(
            url=f"{state_sdk.wb.host}/v1/model-monitors",
            json=model_multi_create_payload,
            headers={"api-key": state_sdk.wb.api_key},
        )

        model_monitor = state_sdk.wb.create_model_monitor(
            model_id="mock_model_id",
            name="test",
            status=MonitorStatus.active,
            metric=MonitorMetrics.accuracy,
            feature="feature1",
            lower_threshold=0.7,
            severity=AlertSeverity.high,
            email="jaclie.chan@chinamail.io",
        )

        assert model_monitor is not None


@pytest.mark.order(get_order_number("sdk_get_drifting_metrics"))
def test_sdk_get_drifting_metrics(client):
    mock_model_id = "mock_model_id"
    with requests_mock.Mocker() as m:
        m.get(
            url=f"{state_sdk.wb.host}/v1/drifting-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_404_NOT_FOUND,
        )

        not_found_result = state_sdk.wb.get_drifting_metrics(model_id=mock_model_id)

        assert not_found_result == None

        m.get(
            url=f"{state_sdk.wb.host}/v1/drifting-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            json=drifting_metrics_report_payload,
        )

        drifting_report = state_sdk.wb.get_drifting_metrics(model_id=mock_model_id)

        assert drifting_report == drifting_metrics_report_payload


@pytest.mark.order(get_order_number("sdk_get_descriptive_statistics"))
def test_sdk_get_descriptive_statistics(client):
    mock_model_id = "mock_model_id"
    with requests_mock.Mocker() as m:
        m.get(
            url=f"{state_sdk.wb.host}/v1/model-integrity-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_404_NOT_FOUND,
        )

        not_found_result = state_sdk.wb.get_descriptive_statistics(
            model_id=mock_model_id
        )

        assert not_found_result == None

        m.get(
            url=f"{state_sdk.wb.host}/v1/model-integrity-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            json=descriptive_statistics_report_payload,
        )

        descriptive_report = state_sdk.wb.get_descriptive_statistics(
            model_id=mock_model_id
        )

        assert descriptive_report == descriptive_statistics_report_payload


@pytest.mark.order(get_order_number("sdk_get_performance_metrics"))
def test_sdk_get_performance_metrics(client):
    mock_model_id = "mock_model_id"
    with requests_mock.Mocker() as m:
        m.get(
            url=f"{state_sdk.wb.host}/v1/performance-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_404_NOT_FOUND,
        )

        not_found_result = state_sdk.wb.get_performance_metrics(model_id=mock_model_id)

        assert not_found_result == None

        m.get(
            url=f"{state_sdk.wb.host}/v1/performance-metrics?model_id={mock_model_id}",
            headers={"api-key": state_sdk.wb.api_key},
            json=performance_metrics_report_payload,
        )

        performance_report = state_sdk.wb.get_performance_metrics(
            model_id=mock_model_id
        )

        assert performance_report == performance_metrics_report_payload


@pytest.mark.order(get_order_number("sdk_get_xai_row"))
def test_sdk_get_xai_row(client):
    mock_inference_id = "mock_inference_id"
    with requests_mock.Mocker() as m:
        m.get(
            url=f"{state_sdk.wb.host}/v1/inference-rows/{mock_inference_id}/xai",
            headers={"api-key": state_sdk.wb.api_key},
            status_code=status.HTTP_404_NOT_FOUND,
        )

        not_found_result = state_sdk.wb.get_xai_row(inference_row_id=mock_inference_id)

        assert not_found_result == None

        m.get(
            url=f"{state_sdk.wb.host}/v1/inference-rows/{mock_inference_id}/xai",
            headers={"api-key": state_sdk.wb.api_key},
            json=inference_row_xai_payload,
        )

        xai = state_sdk.wb.get_xai_row(inference_row_id=mock_inference_id)

        assert xai == inference_row_xai_payload



from typing import Any, List
from whitebox.crud.base import CRUDBase
from sqlalchemy.orm import Session
from whitebox.schemas.modelMonitor import (
    ModelMonitor,
    ModelMonitorCreateDto,
    MonitorStatus,
)
from whitebox.entities.ModelMonitor import ModelMonitor as ModelMonitorEntity


class CRUD(CRUDBase[ModelMonitor, ModelMonitorCreateDto, Any]):
    def get_model_monitors_by_model(
        self, db: Session, *, model_id: str
    ) -> List[ModelMonitor]:
        return (
            db.query(self.model).filter(ModelMonitorEntity.model_id == model_id).all()
        )

    def get_active_model_monitors_by_model(
        self, db: Session, *, model_id: str
    ) -> List[ModelMonitor]:
        return (
            db.query(self.model)
            .filter(
                ModelMonitorEntity.model_id == model_id,
                ModelMonitorEntity.status == MonitorStatus.active,
            )
            .all()
        )


model_monitors = CRUD(ModelMonitorEntity)



import shutil
from fastapi.testclient import TestClient
from pytest import fixture
from sqlalchemy.orm import close_all_sessions
from whitebox import crud
from whitebox.core.settings import get_settings
from whitebox.entities.Base import Base
from whitebox.main import app
from whitebox.sdk.whitebox import Whitebox
from whitebox.tests.utils.maps import v1_test_order_map
from whitebox.utils.passwords import decrypt_api_key
from whitebox.core.db import SessionLocal, engine

settings = get_settings()


def get_order_number(task):
    return v1_test_order_map.index(task)


@fixture(scope="session")
def client():
    with TestClient(app) as client:
        yield client


@fixture(scope="session")
def api_key():
    db = SessionLocal()

    user = crud.users.get_first_by_filter(db=db, username="admin")
    api_key = (
        decrypt_api_key(user.api_key, settings.SECRET_KEY.encode())
        if settings.SECRET_KEY
        else user.api_key
    )

    yield api_key


@fixture(scope="session", autouse=True)
def drop_db():
    yield
    # Removes the folder "test_model" with the models trained during testing
    test_model_path = settings.MODEL_PATH
    shutil.rmtree(test_model_path)

    close_all_sessions()
    # Drops all test database tables
    Base.metadata.drop_all(engine)


class TestsState:
    user: dict = {}
    model_binary: dict = {}
    model_multi: dict = {}
    model_multi_2: dict = {}
    model_multi_3: dict = {}
    model_regression: dict = {}
    inference_row_multi: dict = {}
    inference_row_binary: dict = {}


state = TestsState()


class TestsSDKState:
    wb: Whitebox


state_sdk = TestsSDKState()



from typing import Optional, Union
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase


class UserBase(BaseModel):
    username: str


class User(UserBase, ItemBase):
    pass


class UserCreateDto(UserBase):
    api_key: str



from typing import Dict, Union
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase
from datetime import datetime


class FeatureMetrics(BaseModel):
    missing_count: Dict[str, int]
    non_missing_count: Dict[str, int]
    mean: Dict[str, float]
    minimum: Dict[str, float]
    maximum: Dict[str, float]
    sum: Dict[str, float]
    standard_deviation: Dict[str, float]
    variance: Dict[str, float]


class ModelIntegrityMetricBase(BaseModel):
    model_id: str
    timestamp: Union[str, datetime]
    feature_metrics: FeatureMetrics


class ModelIntegrityMetric(ModelIntegrityMetricBase, ItemBase):
    pass


class ModelIntegrityMetricCreate(ModelIntegrityMetricBase):
    pass



import pytest
from whitebox.tests.v1.conftest import get_order_number
from fastapi import status


@pytest.mark.order(get_order_number("cron_tasks_run_no_models"))
def test_cron_tasks_no_models(client):
    response = client.post("/v1/cron-tasks/run")
    assert response.status_code == status.HTTP_200_OK


@pytest.mark.order(get_order_number("cron_tasks_run_no_inference"))
def test_cron_tasks_no_inference(client):
    response = client.post("/v1/cron-tasks/run")
    assert response.status_code == status.HTTP_200_OK


@pytest.mark.order(get_order_number("cron_tasks_run_ok"))
def test_cron_tasks(client):
    response = client.post("/v1/cron-tasks/run")
    assert response.status_code == status.HTTP_200_OK



from typing import Any, List
from sqlalchemy.orm import Session
from whitebox.crud.base import CRUDBase
from whitebox.schemas.datasetRow import DatasetRow, DatasetRowCreate
from whitebox.entities.DatasetRow import DatasetRow as DatasetRowEntity


class CRUD(CRUDBase[DatasetRow, DatasetRowCreate, Any]):
    def get_dataset_rows_by_model(
        self, db: Session, *, model_id: str
    ) -> List[DatasetRow]:
        return db.query(self.model).filter(DatasetRowEntity.model_id == model_id).all()


dataset_rows = CRUD(DatasetRowEntity)



from sqlalchemy import Column, String, ForeignKey, DateTime, JSON, Float
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid


class InferenceRow(Base):
    __tablename__ = "inference_rows"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    nonprocessed = Column(JSON)
    processed = Column(JSON)
    actual = Column(Float, nullable=True)

    created_at = Column(DateTime)
    updated_at = Column(DateTime)



from whitebox.schemas.utils import HealthCheck
from fastapi import APIRouter, status

health_router = APIRouter()


@health_router.get(
    "/health",
    tags=["Health"],
    response_model=HealthCheck,
    summary="Health check the service",
    status_code=status.HTTP_200_OK,
    response_description="Status of the service",
)
def health_check():
    """Responds with the status of the service."""
    return HealthCheck(status="OK")



import pytest
from whitebox.tests.v1.conftest import get_order_number
from fastapi import status


@pytest.mark.order(get_order_number("models_no_api_key"))
def test_model_no_api_key(client):
    response = client.get("/v1/models")

    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


@pytest.mark.order(get_order_number("models_wrong_api_key"))
def test_model_wrong_api_key(client):
    response = client.get(
        "/v1/models",
        headers={"api-key": "1234567890"},
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED



from whitebox.analytics.metrics.pipelines import *
from whitebox.analytics.drift.pipelines import *
from whitebox.analytics.models.pipelines import *
from whitebox.analytics.xai_models.pipelines import *
from unittest import TestCase
from sklearn.datasets import fetch_california_housing
from sklearn.datasets import load_breast_cancer, load_wine, load_diabetes

test_model_id = "test_model_id"

test_metrics_df = pd.read_csv("whitebox/analytics/data/testing/metrics_test_data.csv")
test_classification_df = pd.read_csv(
    "whitebox/analytics/data/testing/classification_test_data.csv"
)
drift_data = fetch_california_housing(as_frame=True)
drift_data = drift_data.frame
reference = drift_data.head(500)
current = drift_data.iloc[1000:1200]
reference_concept_drift = test_classification_df.head(5)
current_concept_drift = test_classification_df.tail(5)
concept_drift_detected_dataset = pd.read_csv(
    "whitebox/analytics/data/testing/udemy_fin_adj.csv"
)
reference_concept_drift_detected = concept_drift_detected_dataset.head(1000)
current_concept_drift_detected = concept_drift_detected_dataset.tail(1000)
df_load_binary = load_breast_cancer()
df_binary = pd.DataFrame(df_load_binary.data, columns=df_load_binary.feature_names)
df_binary["target"] = df_load_binary.target
df_binary_inference = df_binary.drop(columns=["target"])
df_binary_inference = df_binary_inference.tail(10)
df_binary_inference_row1 = df_binary_inference.iloc[7]
df_binary_inference_row2 = df_binary_inference.iloc[3]
df_load_multi = load_wine()
df_multi = pd.DataFrame(df_load_multi.data, columns=df_load_multi.feature_names)
df_multi["target"] = df_load_multi.target
df_multi_inference = df_multi.drop(columns=["target"])
df_multi_inference = df_multi_inference.tail(10)
df_multi_inference_row1 = df_multi_inference.iloc[4]
df_multi_inference_row2 = df_multi_inference.iloc[2]

test_regression_df = pd.read_csv(
    "whitebox/analytics/data/testing/regression_test_data.csv"
)
df_load_reg = load_diabetes()
df_reg = pd.DataFrame(df_load_reg.data, columns=df_load_reg.feature_names)
df_reg["target"] = df_load_reg.target
df_reg_inference = df_reg.drop(columns=["target"])
df_reg_inference = df_reg_inference.tail(10)
df_reg_inference_row1 = df_reg_inference.iloc[7]
df_reg_inference_row2 = df_reg_inference.iloc[3]


class TestNodes:
    def test_create_feature_metrics_pipeline(self):
        features_metrics = dict(create_feature_metrics_pipeline(test_metrics_df))
        missing_count = features_metrics["missing_count"]
        non_missing_count = features_metrics["non_missing_count"]
        mean = features_metrics["mean"]
        minimum = features_metrics["minimum"]
        maximum = features_metrics["maximum"]
        sum = features_metrics["sum"]
        standard_deviation = features_metrics["standard_deviation"]
        variance = features_metrics["variance"]
        TestCase().assertDictEqual(
            {"num1": 1, "num2": 2, "num3": 0, "cat1": 1, "cat2": 2}, missing_count
        )
        TestCase().assertDictEqual(
            {"num1": 9, "num2": 8, "num3": 10, "cat1": 9, "cat2": 8}, non_missing_count
        )
        TestCase().assertDictEqual(
            {"num1": 156.33333333333334, "num2": 9.223817500000001, "num3": 1.0}, mean
        )
        TestCase().assertDictEqual({"num1": 0.0, "num2": 0.00054, "num3": 0.0}, minimum)
        TestCase().assertDictEqual(
            {"num1": 1000.0, "num2": 45.896, "num3": 2.0}, maximum
        )
        TestCase().assertDictEqual(
            {"num1": 1407.0, "num2": 73.79054000000001, "num3": 10.0}, sum
        )
        TestCase().assertDictEqual(
            {
                "num1": 322.0283372624217,
                "num2": 15.488918075768835,
                "num3": 0.816496580927726,
            },
            standard_deviation,
        )
        TestCase().assertDictEqual(
            {
                "num1": 103702.25000000001,
                "num2": 239.90658315787854,
                "num3": 0.6666666666666666,
            },
            variance,
        )

    def test_create_binary_classification_evaluation_metrics_pipeline(self):
        binary_metrics = dict(
            create_binary_classification_evaluation_metrics_pipeline(
                test_classification_df["y_testing_binary"],
                test_classification_df["y_prediction_binary"],
                labels=[0, 1],
            )
        )

        binary_metrics_edge_case = dict(
            create_binary_classification_evaluation_metrics_pipeline(
                test_classification_df["y_testing_binary"].tail(1),
                test_classification_df["y_prediction_binary"].tail(1),
                labels=[0, 1],
            )
        )

        assert binary_metrics_edge_case["true_positive"] == 1
        assert binary_metrics["accuracy"] == 0.6
        assert binary_metrics["precision"] == 0.6
        assert binary_metrics["recall"] == 0.6
        assert binary_metrics["f1"] == 0.6
        assert binary_metrics["true_negative"] == 3
        assert binary_metrics["false_positive"] == 2
        assert binary_metrics["false_negative"] == 2
        assert binary_metrics["true_positive"] == 3

    def test_create_multiple_classification_evaluation_metrics_pipeline(self):
        multi_metrics = create_multiple_classification_evaluation_metrics_pipeline(
            test_classification_df["y_testing_multi"],
            test_classification_df["y_prediction_multi"],
            labels=[0, 1, 2],
        )

        multi_metrics_edge_case = (
            create_multiple_classification_evaluation_metrics_pipeline(
                test_classification_df["y_testing_multi"].tail(1),
                test_classification_df["y_prediction_multi"].tail(1),
                labels=[0, 1, 2],
            )
        )

        assert multi_metrics.accuracy == 0.6

        TestCase().assertDictEqual(
            {
                "true_negative": 0,
                "false_positive": 0,
                "false_negative": 0,
                "true_positive": 1,
            },
            multi_metrics_edge_case.confusion_matrix["class1"].dict(),
        )

        TestCase().assertDictEqual(
            {"micro": 0.6, "macro": 0.6444444444444445, "weighted": 0.64},
            multi_metrics.precision.dict(),
        )
        TestCase().assertDictEqual(
            {"micro": 0.6, "macro": 0.5833333333333334, "weighted": 0.6},
            multi_metrics.recall.dict(),
        )
        TestCase().assertDictEqual(
            {"micro": 0.6, "macro": 0.6, "weighted": 0.6066666666666667},
            multi_metrics.f1.dict(),
        )
        TestCase().assertDictEqual(
            {
                "true_negative": 5,
                "false_positive": 2,
                "false_negative": 2,
                "true_positive": 1,
            },
            multi_metrics.confusion_matrix["class0"].dict(),
        )
        TestCase().assertDictEqual(
            {
                "true_negative": 4,
                "false_positive": 2,
                "false_negative": 1,
                "true_positive": 3,
            },
            multi_metrics.confusion_matrix["class1"].dict(),
        )
        TestCase().assertDictEqual(
            {
                "true_negative": 7,
                "false_positive": 0,
                "false_negative": 1,
                "true_positive": 2,
            },
            multi_metrics.confusion_matrix["class2"].dict(),
        )

    def test_create_regression_evaluation_metrics_pipeline(self):
        reg_report = dict(
            create_regression_evaluation_metrics_pipeline(
                test_regression_df["y_test"], test_regression_df["y_prediction"]
            )
        )
        rsq = reg_report["r_square"]
        mse = reg_report["mean_squared_error"]
        mae = reg_report["mean_absolute_error"]
        assert (rsq) == 0.9044
        assert (mse) == 0.0071
        assert (mae) == 0.037

    def test_create_data_drift_pipeline(self):
        data_drift_report = dict(run_data_drift_pipeline(reference, current))
        assert data_drift_report["number_of_columns"] == 9
        assert data_drift_report["number_of_drifted_columns"] == 7
        assert (
            round(
                dict(data_drift_report["drift_by_columns"]["Population"])[
                    "drift_score"
                ],
                2,
            )
            == 0.06
        )
        assert (
            dict(data_drift_report["drift_by_columns"]["Longitude"])["drift_detected"]
            == True
        )
        assert (
            dict(data_drift_report["drift_by_columns"]["AveBedrms"])["drift_detected"]
            == False
        )

    def test_create_concept_drift_pipeline_drift_not_detected(self):
        concept_drift_report = vars(
            run_concept_drift_pipeline(
                reference_concept_drift, current_concept_drift, "y_testing_multi"
            )
        )
        assert list(concept_drift_report.keys()) == [
            "concept_drift_summary",
            "column_correlation",
        ]
        assert (
            round(vars(concept_drift_report["concept_drift_summary"])["drift_score"], 3)
            == 0.082
        )
        assert (
            vars(concept_drift_report["concept_drift_summary"])["drift_detected"]
            == False
        )

    def test_create_concept_drift_pipeline_drift_detected(self):
        concept_drift_report = vars(
            run_concept_drift_pipeline(
                reference_concept_drift_detected,
                current_concept_drift_detected,
                "discount_price__currency",
            )
        )
        assert list(concept_drift_report.keys()) == [
            "concept_drift_summary",
            "column_correlation",
        ]
        assert (
            round(vars(concept_drift_report["concept_drift_summary"])["drift_score"], 3)
            == 0.008
        )
        assert (
            vars(concept_drift_report["concept_drift_summary"])["drift_detected"]
            == True
        )

    def test_create_binary_classification_training_model_pipeline(self):
        model, eval = create_binary_classification_training_model_pipeline(
            df_binary, "target", test_model_id
        )
        eval_score = eval["roc_auc_score"]
        assert (round(eval_score, 3)) == 0.976

    def test_create_multiclass_classification_training_model_pipeline(self):
        model, eval = create_multiclass_classification_training_model_pipeline(
            df_multi, "target", test_model_id
        )
        eval_score = eval["precision"]
        assert (round(eval_score, 2)) == 0.96

    def test_create_regression_training_model_pipeline(self):
        model, eval = create_regression_training_model_pipeline(
            df_reg, "target", test_model_id
        )
        eval_score = eval["r2_score"]
        assert (eval_score) == 0.2576

    def test_create_xai_pipeline_per_inference_row(self):
        binary_class_report1 = create_xai_pipeline_per_inference_row(
            df_binary, "target", df_binary_inference_row1, "binary", test_model_id
        )
        multi_class_report1 = create_xai_pipeline_per_inference_row(
            df_multi, "target", df_multi_inference_row1, "multi_class", test_model_id
        )
        regression_report1 = create_xai_pipeline_per_inference_row(
            df_reg, "target", df_reg_inference_row1, "regression", test_model_id
        )
        binary_class_report2 = create_xai_pipeline_per_inference_row(
            df_binary, "target", df_binary_inference_row2, "binary", test_model_id
        )
        multi_class_report2 = create_xai_pipeline_per_inference_row(
            df_multi, "target", df_multi_inference_row2, "multi_class", test_model_id
        )
        regression_report2 = create_xai_pipeline_per_inference_row(
            df_reg, "target", df_reg_inference_row2, "regression", test_model_id
        )

        binary_contribution_check_one = binary_class_report1["worst perimeter"]
        binary_contribution_check_two = binary_class_report2["worst texture"]
        multi_contribution_check_one = multi_class_report1["hue"]
        multi_contribution_check_two = multi_class_report2["alcohol"]
        regression_contribution_check_one = regression_report1["sex"]
        regression_contribution_check_two = regression_report2["bp"]

        assert (round(binary_contribution_check_one, 3)) == -0.464
        assert (round(binary_contribution_check_two, 1)) == -0.1
        assert (round(multi_contribution_check_one, 2)) == -0.09
        assert (round(multi_contribution_check_two, 3)) == 0.076
        assert (round(regression_contribution_check_one, 2)) == 9.48
        assert (round(regression_contribution_check_two, 3)) == 14.079



from datetime import datetime
from whitebox.schemas.model import ModelType
from sklearn.datasets import load_breast_cancer, load_wine, load_diabetes
import pandas as pd
import random
from copy import deepcopy

from whitebox.schemas.modelMonitor import AlertSeverity, MonitorMetrics, MonitorStatus

user_create_payload = dict(username="admin")

model_binary_create_payload = dict(
    name="Model 1",
    description="Model 1 description",
    type=ModelType.binary,
    labels={"label_1": 0, "label_2": 1},
    prediction="target",
)

model_multi_create_payload = dict(
    name="Model 2",
    description="Model 2 description",
    type=ModelType.multi_class,
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
)

model_multi_2_create_payload = dict(
    name="Model 3",
    description="Model 3 description",
    type=ModelType.multi_class,
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
)

model_multi_3_create_payload = dict(
    name="Model 4",
    description="Model 4 description",
    type=ModelType.multi_class,
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
)

model_regression_create_payload = dict(
    name="Regression Model",
    description="Regression Model description",
    type=ModelType.regression,
    prediction="target",
)

model_update_payload = dict(
    name="Model 1 - categorical",
    description="Model 1 description",
)

# dataset rows data for both binary and multiclass models
dataset_rows_single_row_column_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0, "target": 1},
    }
]

dataset_rows_no_prediction_column_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0},
    },
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 1, "additionalProp2": 0},
    },
]

dataset_rows_one_prediction_value_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0, "target": 0},
    },
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 1, "additionalProp2": 0, "target": 0},
    },
]


df_load_multi = load_wine()
df_multi = pd.DataFrame(df_load_multi.data, columns=df_load_multi.feature_names)
df_multi["target"] = df_load_multi.target
df_multi = df_multi.tail(100)

dict_multi_data = df_multi.to_dict(orient="records")
dataset_rows_create_multi_class_payload = [
    {"processed": x, "nonprocessed": x} for x in dict_multi_data
]

df_load_binary = load_breast_cancer()
df_binary = pd.DataFrame(df_load_binary.data, columns=df_load_binary.feature_names)
df_binary["target"] = df_load_binary.target
df_binary = df_binary.tail(100)

dict_binary_data = df_binary.to_dict(orient="records")
dataset_rows_create_binary_payload = [
    {"processed": x, "nonprocessed": x} for x in dict_binary_data
]

df_load_reg = load_diabetes()
df_reg = pd.DataFrame(df_load_reg.data, columns=df_load_reg.feature_names)
df_reg["target"] = df_load_reg.target
df_reg = df_reg.tail(100)

dict_reg_data = df_reg.to_dict(orient="records")
dataset_rows_create_reg_payload = [
    {"processed": x, "nonprocessed": x} for x in dict_reg_data
]

dataset_rows_create_wrong_model_payload = list(
    (
        dict(
            model_id="wrong_model_id",
            nonprocessed={"sex": "male"},
            processed={"sex": 0},
        ),
        dict(
            model_id="wrong_model_id",
            nonprocessed={"sex": "female"},
            processed={"sex": 1},
        ),
    )
)


# inference rows data for both binary and multiclaas models
df_multi_inference = df_multi.tail(10)
dict_multi_inferences = df_multi_inference.to_dict(orient="records")
inference_row_create_many_multi_payload = [
    {
        "timestamp": str(datetime.now()),
        "processed": x,
        "nonprocessed": x,
        "actual": random.randint(0, 1),
    }
    for x in dict_multi_inferences
]

inference_row_create_single_row_payload = inference_row_create_many_multi_payload[0]
inference_row_create_many_multi_no_actual_payload = deepcopy(
    inference_row_create_many_multi_payload
)
for x in inference_row_create_many_multi_no_actual_payload:
    del x["actual"]

inference_row_create_many_multi_mixed_actuals_payload = (
    inference_row_create_many_multi_no_actual_payload
    + inference_row_create_many_multi_payload
)

# This is the body of the request coming from the sdk
df_binary_inference = df_binary.tail(10)
dict_binary_inferences = df_binary_inference.to_dict(orient="records")
inference_row_create_many_binary_payload = [
    {
        "timestamp": str(datetime.now()),
        "processed": x,
        "nonprocessed": x,
        "actual": random.randint(0, 1),
    }
    for x in dict_binary_inferences
]

df_reg_inference = df_reg.tail(10)
dict_reg_inferences = df_reg_inference.to_dict(orient="records")
inference_row_create_many_reg_payload = [
    {
        "timestamp": str(datetime.now()),
        "processed": x,
        "nonprocessed": x,
        "actual": random.randint(0, 1),
    }
    for x in dict_reg_inferences
]

timestamps = pd.Series(["2022-12-22T12:13:27.879738"] * 10)
mixed_actuals = pd.Series([0, 1, None, 1, 0, None, None, 1, 0, None])

model_monitor_accuracy_create_payload = dict(
    name="accuracy monitor ",
    status=MonitorStatus.active,
    metric=MonitorMetrics.accuracy,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_f1_create_payload = dict(
    name="f1 Monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.f1,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_data_drift_create_payload = dict(
    name="data drift monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.data_drift,
    feature="concavity error",
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_concept_drift_create_payload = dict(
    name="concept drift monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.concept_drift,
    feature="concavity error",
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_precision_create_payload = dict(
    name="precision monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.precision,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_r_square_create_payload = dict(
    name="r_square monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.r_square,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

performance_metrics_report_payload = [
    {
        "id": "f7030044-e0c2-4493-8aea-f66e6564efb7",
        "created_at": "2023-02-06T09:53:00.343616",
        "updated_at": "2023-02-06T09:53:00.343616",
        "model_id": "dbb7f384-cf46-4884-bca3-7787e049a3d5",
        "timestamp": "2023-02-06T09:53:00.343548",
        "accuracy": 0.6,
        "precision": 0.25,
        "recall": 0.5,
        "f1": 0.3333333333333333,
        "true_negative": 5,
        "false_positive": 3,
        "false_negative": 1,
        "true_positive": 1,
    }
]

descriptive_statistics_report_payload = [
    {
        "id": "d4f18837-65f6-45c2-af7c-34dcdc4aabf9",
        "created_at": "2023-02-06T09:53:00.370639",
        "updated_at": "2023-02-06T09:53:00.370639",
        "model_id": "bc01a212-b01e-48b4-aeb5-4a0a3cfa3c98",
        "timestamp": "2023-02-06T09:53:00.370617",
        "feature_metrics": {
            "missing_count": {
                "alcohol": 0,
                "malic_acid": 0,
                "ash": 0,
                "alcalinity_of_ash": 0,
                "magnesium": 0,
                "total_phenols": 0,
                "flavanoids": 0,
                "nonflavanoid_phenols": 0,
                "proanthocyanins": 0,
                "color_intensity": 0,
                "hue": 0,
                "od280/od315_of_diluted_wines": 0,
                "proline": 0,
                "target": 0,
            },
            "non_missing_count": {
                "alcohol": 10,
                "malic_acid": 10,
                "ash": 10,
                "alcalinity_of_ash": 10,
                "magnesium": 10,
                "total_phenols": 10,
                "flavanoids": 10,
                "nonflavanoid_phenols": 10,
                "proanthocyanins": 10,
                "color_intensity": 10,
                "hue": 10,
                "od280/od315_of_diluted_wines": 10,
                "proline": 10,
                "target": 10,
            },
            "mean": {
                "alcohol": 13.379000000000001,
                "malic_acid": 3.564,
                "ash": 2.493,
                "alcalinity_of_ash": 21.6,
                "magnesium": 102.3,
                "total_phenols": 1.6620000000000001,
                "flavanoids": 0.699,
                "nonflavanoid_phenols": 0.44499999999999995,
                "proanthocyanins": 1.1889999999999998,
                "color_intensity": 8.595999899999999,
                "hue": 0.6399999999999999,
                "od280/od315_of_diluted_wines": 1.6970000000000003,
                "proline": 674.5,
                "target": 2.0,
            },
            "minimum": {
                "alcohol": 12.2,
                "malic_acid": 2.39,
                "ash": 2.26,
                "alcalinity_of_ash": 19.0,
                "magnesium": 86.0,
                "total_phenols": 1.25,
                "flavanoids": 0.49,
                "nonflavanoid_phenols": 0.27,
                "proanthocyanins": 0.64,
                "color_intensity": 5.5,
                "hue": 0.57,
                "od280/od315_of_diluted_wines": 1.56,
                "proline": 470.0,
                "target": 2.0,
            },
            "maximum": {
                "alcohol": 14.16,
                "malic_acid": 5.65,
                "ash": 2.86,
                "alcalinity_of_ash": 25.0,
                "magnesium": 120.0,
                "total_phenols": 2.05,
                "flavanoids": 0.96,
                "nonflavanoid_phenols": 0.56,
                "proanthocyanins": 1.54,
                "color_intensity": 10.2,
                "hue": 0.74,
                "od280/od315_of_diluted_wines": 1.92,
                "proline": 840.0,
                "target": 2.0,
            },
            "sum": {
                "alcohol": 133.79000000000002,
                "malic_acid": 35.64,
                "ash": 24.93,
                "alcalinity_of_ash": 216.0,
                "magnesium": 1023.0,
                "total_phenols": 16.62,
                "flavanoids": 6.989999999999999,
                "nonflavanoid_phenols": 4.449999999999999,
                "proanthocyanins": 11.889999999999999,
                "color_intensity": 85.959999,
                "hue": 6.3999999999999995,
                "od280/od315_of_diluted_wines": 16.970000000000002,
                "proline": 6745.0,
                "target": 20.0,
            },
            "standard_deviation": {
                "alcohol": 0.5907894906159238,
                "malic_acid": 1.1073311258256142,
                "ash": 0.2058613341278272,
                "alcalinity_of_ash": 2.3664319132398464,
                "magnesium": 11.80442478244681,
                "total_phenols": 0.24334703157790474,
                "flavanoids": 0.14224001624796806,
                "nonflavanoid_phenols": 0.08396427811873335,
                "proanthocyanins": 0.3045743842734572,
                "color_intensity": 1.4311393049673125,
                "hue": 0.05291502622129182,
                "od280/od315_of_diluted_wines": 0.12356284950492914,
                "proline": 130.39363481397396,
                "target": 0.0,
            },
            "variance": {
                "alcohol": 0.34903222222222274,
                "malic_acid": 1.2261822222222223,
                "ash": 0.04237888888888891,
                "alcalinity_of_ash": 5.599999999999999,
                "magnesium": 139.34444444444443,
                "total_phenols": 0.059217777777777765,
                "flavanoids": 0.02023222222222222,
                "nonflavanoid_phenols": 0.0070500000000000024,
                "proanthocyanins": 0.09276555555555556,
                "color_intensity": 2.048159710222322,
                "hue": 0.002800000000000001,
                "od280/od315_of_diluted_wines": 0.015267777777777769,
                "proline": 17002.5,
                "target": 0.0,
            },
        },
    }
]

drifting_metrics_report_payload = [
    {
        "id": "d11f2313-12e3-4533-b16d-7689175ad617",
        "created_at": "2023-02-06T09:53:00.450466",
        "updated_at": "2023-02-06T09:53:00.450466",
        "model_id": "e8e5ff00-0212-4a1e-bcdb-0ef9183c867f",
        "timestamp": "2023-02-06T09:53:00.450435",
        "concept_drift_summary": {
            "concept_drift_summary": {
                "column_name": "target",
                "column_type": "cat",
                "stattest_name": "Z-test p_value",
                "drift_score": 0.0010364142456651404,
                "drift_detected": True,
                "stattest_threshold": 0.05,
            },
            "column_correlation": {
                "column_name": "target",
                "current": {},
                "reference": {},
            },
        },
        "data_drift_summary": {
            "number_of_columns": 13,
            "number_of_drifted_columns": 7,
            "share_of_drifted_columns": 0.5384615384615384,
            "dataset_drift": True,
            "drift_by_columns": {
                "alcalinity_of_ash": {
                    "column_name": "alcalinity_of_ash",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.5163177458201748,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
                "alcohol": {
                    "column_name": "alcohol",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.0012612112534500996,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "ash": {
                    "column_name": "ash",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.23438597334285016,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
                "color_intensity": {
                    "column_name": "color_intensity",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.00015242693342585546,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "flavanoids": {
                    "column_name": "flavanoids",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.0003384276858653337,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "hue": {
                    "column_name": "hue",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.0002500181491633883,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "magnesium": {
                    "column_name": "magnesium",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.08450859981880006,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
                "malic_acid": {
                    "column_name": "malic_acid",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.033698127881912385,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "nonflavanoid_phenols": {
                    "column_name": "nonflavanoid_phenols",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.30861039864383094,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
                "od280/od315_of_diluted_wines": {
                    "column_name": "od280/od315_of_diluted_wines",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.0006076739224878145,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "proanthocyanins": {
                    "column_name": "proanthocyanins",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.09673060972050829,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
                "proline": {
                    "column_name": "proline",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.0328848734766648,
                    "drift_detected": True,
                    "threshold": 0.05,
                },
                "total_phenols": {
                    "column_name": "total_phenols",
                    "column_type": "num",
                    "stattest_name": "K-S p_value",
                    "drift_score": 0.082145120769578,
                    "drift_detected": False,
                    "threshold": 0.05,
                },
            },
        },
    }
]

inference_row_xai_payload = {
    "mean perimeter": -0.23641989216827916,
    "mean concavity": -0.16854638344893477,
    "worst radius": -0.16657884627852115,
    "mean concave points": -0.1260514980862101,
    "worst perimeter": -0.08533726601274261,
    "worst texture": -0.07905090544100012,
    "mean texture": -0.05969642438812264,
    "worst concave points": -0.03310756238929911,
    "mean radius": -0.03244286722644983,
    "worst symmetry": 0.02852296352595751,
}



import asyncio
import functools
from collections import deque

from crontab import CronTab
from whitebox.utils.exceptions import (
    TaskNotFoundException,
    TaskAlreadyRunningException,
    TaskNotRunningException,
)
from whitebox.utils.logger import cronLogger as logger
from whitebox.schemas.task import (
    TaskDefinition,
    RunningTask,
    TaskLog,
    TaskInfo,
    TaskStatus,
    State,
    TaskRealTimeInfo,
    EventType,
)
from typing import Callable, Optional, Coroutine, List, Dict, Deque, Union

from functools import lru_cache
import datetime

import pytz


def now():
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


class Task_Manager:
    def __init__(self):
        self._definitions: Dict[str, TaskDefinition] = {}
        self._real_time: Dict[str, TaskRealTimeInfo] = {}
        self._running_tasks: Dict[str, RunningTask] = {}
        self._log_queue: Deque[TaskLog] = deque()

        self._is_running: bool = False
        self._is_shutting_down: bool = False

        self._cleanup_tasks: List[asyncio.Task] = []

    def clear(self):
        """Resets the"""
        if self._is_running or self._is_shutting_down:
            raise Exception("Cannot clear before shutdown")
        self._definitions: Dict[str, TaskDefinition] = {}
        self._real_time: Dict[str, TaskRealTimeInfo] = {}
        self._running_tasks: Dict[str, RunningTask] = {}
        self._log_queue: Deque[TaskLog] = deque()

    def register(
        self,
        async_callable: Callable[[], Coroutine],
        crontab: Union[str, None] = None,
        name: Union[str, None] = None,
    ):
        name = name or async_callable.__name__
        if name in self._definitions:
            raise Exception(f"Task <{name}> already exists.")

        definition = TaskDefinition(
            name=name, async_callable=async_callable, crontab=crontab, enabled=True
        )
        self._definitions[name] = definition
        self._real_time[name] = TaskRealTimeInfo(
            name=name, status="registered", next_run_ts=None
        )

        self._log_event("task_registered", definition.name)

    def _log_event(
        self, event_type: EventType, task_name: str, error: Union[str, None] = None
    ):
        self._log_queue.append(
            TaskLog(
                event_type=event_type,
                task_name=task_name,
                crontab=self._definitions[task_name].crontab,
                enabled=self._definitions[task_name].enabled,
                error=error,
            )
        )

    def _get_task_definition(self, name: str) -> TaskDefinition:
        try:
            return self._definitions[name]
        except KeyError as e:
            raise TaskNotFoundException from e

    def _get_running_task(self, name: str) -> RunningTask:
        try:
            return self._running_tasks[name]
        except KeyError as e:
            raise TaskNotRunningException from e

    def _is_task_running(self, name: str) -> bool:
        return name in self._running_tasks

    def get_task_info(self, name: str) -> TaskInfo:
        definition = self._get_task_definition(name)
        return TaskInfo(
            name=definition.name,
            enabled=definition.enabled,
            crontab=definition.crontab,
            started_at=self._get_task_started_at(name),
            stopped_at=self._get_task_stopped_at(name),
            next_run_in=self._get_task_next_run_in(name),
            previous_status=self._get_previous_status(name) or "registered",
            status=self._get_task_status(name),
        )

    # TODO: Implement functionality
    def _get_task_started_at(self, name: str):
        rt_info = self._real_time.get(name)
        if not rt_info:
            raise TaskNotFoundException("task not found")
        return rt_info.started_at

    # TODO: Implement functionality
    def _get_task_stopped_at(self, name: str):
        rt_info = self._real_time.get(name)
        if not rt_info:
            raise TaskNotFoundException("task not found")
        return rt_info.stopped_at

    def _get_task_status(self, name: str) -> TaskStatus:
        rt_info = self._real_time.get(name)
        if not rt_info:
            raise TaskNotFoundException("task not found")
        return rt_info.status

    def _get_previous_status(self, name: str) -> TaskStatus:
        rt_info = self._real_time.get(name)
        if not rt_info:
            raise TaskNotFoundException("task not found")
        return rt_info.previous_status

    def _get_task_next_run_in(self, name: str) -> Optional[int]:
        definition = self._get_task_definition(name)
        if definition.crontab is None:
            return None
        return CronTab(definition.crontab).next(default_utc=True)

    def _create_running_task(self, definition: TaskDefinition) -> RunningTask:
        running_task = RunningTask(
            task_definition=definition,
            asyncio_task=asyncio.get_event_loop().create_task(
                definition.async_callable()
            ),
            since=now().timestamp(),
        )
        running_task.asyncio_task.add_done_callback(
            functools.partial(self._on_task_done, definition.name)
        )
        return running_task

    async def cancel_task(self, name: str):
        if not self._is_task_running(name):
            raise TaskNotRunningException("task not running")
        logger.info(f"Cancelling {name}")
        self._get_task_definition(name)
        running_task = self._get_running_task(name)
        cancelled = running_task.asyncio_task.cancel()
        return cancelled

    def run_task(self, name: str) -> None:
        if self._is_task_running(name):
            raise TaskAlreadyRunningException("task already running")

        definition = self._get_task_definition(name)
        running_task = self._create_running_task(definition)
        self._running_tasks[definition.name] = running_task

        self._real_time[name] = TaskRealTimeInfo(
            name=name,
            status="running",
            previous_status=self._get_previous_status(name),
            started_at=now().timestamp(),
            next_run_ts=now().timestamp() + (self._get_task_next_run_in(name) or 0),
        )

    # TODO: Implement disable task functionality
    def disable_task(self, name: str):
        return False

    # TODO: Implement enable task functionality
    def enable_task(self, name: str):
        return False

    def get_all_tasks_info(self) -> List[TaskInfo]:
        return [self.get_task_info(name) for name in self._definitions.keys()]

    def _on_task_done(self, task_name: str, task: asyncio.Task) -> None:
        definition = self._get_task_definition(task_name)
        del self._running_tasks[task_name]

        try:
            exception = task.exception()
        except asyncio.CancelledError:
            self._log_event("task_cancelled", definition.name)

            self._real_time[task_name] = TaskRealTimeInfo(
                name=task_name,
                status="pending",
                previous_status="cancelled",
                next_run_ts=None,
                started_at=self._get_task_started_at(task_name),
                stopped_at=now().timestamp(),
            )
            task = asyncio.get_event_loop().create_task(
                self.on_task_cancelled(task_name)
            )
            self._cleanup_tasks.append(task)
            return

        if exception:
            self._log_event("task_failed", definition.name, error=str(exception))

            self._real_time[task_name] = TaskRealTimeInfo(
                name=task_name,
                status="pending",
                previous_status="failed",
                next_run_ts=None,
                started_at=self._get_task_started_at(task_name),
                stopped_at=now().timestamp(),
            )
            task = asyncio.get_event_loop().create_task(
                self.on_task_exception(task_name, exception)
            )
            self._cleanup_tasks.append(task)
            return

        self._log_event("task_finished", definition.name)

        self._real_time[task_name] = TaskRealTimeInfo(
            name=task_name,
            status="pending",
            previous_status="finished",
            started_at=self._get_task_started_at(task_name),
            stopped_at=now().timestamp(),
            next_run_ts=now().timestamp() + self._get_task_next_run_in(task_name)
            if definition.crontab
            else None,
        )
        task = asyncio.get_event_loop().create_task(self.on_task_finished(task_name))
        self._cleanup_tasks.append(task)
        return

    async def _on_task_started(self, task_name: str):
        definition = self._get_task_definition(task_name)

        self._log_event("task_started", definition.name)
        await self.on_task_started(task_name)

    async def run(self, state: Union[State, None] = None):
        if self._is_running:
            logger.warning("Ignoring current calling of run(). Already running.")
            return

        self._is_running = True
        await self.on_startup()

        if state:
            for task_info in state.tasks_info:
                if task_info.name in self._definitions:
                    self._definitions[task_info.name].crontab = task_info.crontab
                    self._definitions[task_info.name].enabled = task_info.enabled

                    self._real_time[task_info.name].status = task_info.status

        await self._run_ad_infinitum()

    async def _run_ad_infinitum(self):
        while True and self._is_running:
            for task_name, rt_info in self._real_time.items():
                this_time_ts = now().timestamp()

                if rt_info.status == "registered":
                    delta: int = self._get_task_next_run_in(task_name) or 0

                    self._real_time[task_name] = TaskRealTimeInfo(
                        name=task_name,
                        status="pending",
                        previous_status="registered",
                        next_run_ts=now().timestamp() + delta,
                    )
                elif (
                    not self._is_shutting_down
                    and rt_info.status in ["pending", "finished"]
                    and rt_info.next_run_ts is not None
                    and rt_info.next_run_ts <= this_time_ts
                ):
                    self.run_task(task_name)
                elif rt_info.status == "running" and not self._is_task_running(
                    task_name
                ):
                    self.run_task(task_name)
                else:  # rt_info.status in ["cancelled", "failed"]:
                    ...
            await asyncio.sleep(1.5)

    async def shutdown(self):
        await asyncio.sleep(2)
        logger.info("Shutting down...")
        logger.info(f"Cancelling {len(self._running_tasks)} running tasks...")
        self._is_shutting_down = True

        for running_task in self._running_tasks.values():
            await self.cancel_task(running_task.task_definition.name)
            logger.debug(f"Cancelled task {running_task.task_definition.name}")

        await asyncio.gather(*self._cleanup_tasks)
        logger.debug("Cleanup tasks finished.")

        await self.on_shutdown()

        self._is_running = False
        self._is_shutting_down = False

    async def on_task_started(self, task_name: str):
        ...

    async def on_task_exception(self, task_name: str, exception: BaseException):
        ...

    async def on_task_cancelled(self, task_name: str):
        logger.info(f"Cancelled {task_name}")

    async def on_task_finished(self, task_name: str):
        logger.info(f"Finished {task_name}")

    async def on_startup(self):
        ...

    async def on_shutdown(self):
        ...

    def state(self) -> State:
        state = State(created_at=now(), tasks_info=self.get_all_tasks_info())
        return state


@lru_cache()
def get_task_manager():
    return Task_Manager()



from whitebox.core.settings import get_settings
import databases
import sqlalchemy
from sqlalchemy.orm import sessionmaker
from whitebox.entities.Base import Base
from whitebox.schemas.user import UserCreateDto

from whitebox import crud
from whitebox.utils.passwords import encrypt_api_key
from whitebox.utils.logger import cronLogger as logger

from secrets import token_hex

settings = get_settings()
database = databases.Database(settings.DATABASE_URL)
engine = sqlalchemy.create_engine(settings.DATABASE_URL)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


async def connect():
    """
    Connect to DB
    """
    Base.metadata.create_all(engine)
    db = SessionLocal()

    admin_exists = crud.users.get_first_by_filter(db=db, username="admin")
    if not admin_exists:
        plain_api_key = token_hex(32)
        secret_key = settings.SECRET_KEY
        api_key = (
            encrypt_api_key(plain_api_key, secret_key.encode())
            if secret_key
            else plain_api_key
        )

        obj_in = UserCreateDto(username="admin", api_key=api_key)
        crud.users.create(db=db, obj_in=obj_in)
        logger.info(f"Created username: admin, API key: {plain_api_key}")
    await database.connect()


async def close():
    """
    Close DB Connection
    """
    await database.disconnect()
    # logging.info("Closed connection with DB")



import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number
from fastapi import status


@pytest.mark.order(get_order_number("health"))
def test_health(client):
    response = client.get("/v1/health")
    assert response.status_code == status.HTTP_200_OK
    validated = schemas.HealthCheck(**response.json())



from typing import List, Union
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.alert import Alert
from whitebox.schemas.user import User
from whitebox.utils.errors import add_error_responses, errors


alerts_router = APIRouter()


@alerts_router.get(
    "/alerts",
    tags=["Alerts"],
    response_model=List[Alert],
    summary="Get all model's alerts",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_alerts(
    model_id: Union[str, None] = None,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """
    Fetches alerts from the databse.
    \n If a model id is provided, only the alerts for the specific model will be fetched.
    \n If a model id is not provided then all alerts from the database will be fetched.
    """

    if model_id:
        model = crud.models.get(db, model_id)
        if model:
            return crud.alerts.get_model_alerts_by_model(db=db, model_id=model_id)
        else:
            return errors.not_found("Model not found")
    else:
        return crud.alerts.get_all(db=db)



from whitebox.crud.base import CRUDBase
from whitebox.schemas.model import Model, ModelCreateDto, ModelUpdateDto
from whitebox.entities.Model import Model as ModelEntity


class CRUD(CRUDBase[Model, ModelCreateDto, ModelUpdateDto]):
    pass


models = CRUD(ModelEntity)



from sqlalchemy import Column, Numeric, Enum, String, ForeignKey, DateTime
from whitebox.entities.Base import Base
from sqlalchemy.orm import relationship
from whitebox.schemas.modelMonitor import AlertSeverity, MonitorMetrics, MonitorStatus
from whitebox.utils.id_gen import generate_uuid


class ModelMonitor(Base):
    __tablename__ = "model_monitors"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    name = Column(String)
    status = Column("status", Enum(MonitorStatus))
    metric = Column("metric", Enum(MonitorMetrics))
    feature = Column(String, nullable=True)
    lower_threshold = Column(Numeric, nullable=True)
    severity = Column("severity", Enum(AlertSeverity))
    email = Column(String)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)

    alerts = relationship("Alert")



from sqlalchemy import Column, String, DateTime
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid
from sqlalchemy.orm import deferred, relationship


class User(Base):
    __tablename__ = "users"

    id = Column(String, unique=True, primary_key=True, default=generate_uuid)
    username = Column(String)
    api_key = Column(String)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)



from datetime import datetime
from typing import Union
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase


class AlertBase(BaseModel):
    model_id: str
    model_monitor_id: str
    timestamp: Union[str, datetime]
    description: str


class Alert(AlertBase, ItemBase):
    pass


class AlertCreateDto(AlertBase):
    pass



from colorama import Fore, Back, Style
import logging


class Logger:
    def info(self, info):
        print(Fore.LIGHTBLUE_EX + "INFO" + Fore.BLACK + ":" + 4 * " ", info)

    def error(self, error):
        print(Fore.RED + "ERROR" + Fore.BLACK + ":" + 3 * " ", error)

    def success(self, msg):
        print(Fore.GREEN + "SUCCESS" + Fore.BLACK + ":" + 1 * " ", msg)


log = Logger()


logging.basicConfig(level="INFO")

cronLogger = logging.getLogger("cron")



from typing import Any
from whitebox.crud.base import CRUDBase
from whitebox.schemas.user import User, UserCreateDto
from whitebox.entities.User import User as UserEntity


class CRUD(CRUDBase[User, UserCreateDto, Any]):
    pass


users = CRUD(UserEntity)



from whitebox.tests.v1.mock_data import (
    inference_row_create_single_row_payload,
    inference_row_create_many_binary_payload,
    inference_row_create_many_multi_payload,
    inference_row_create_many_multi_no_actual_payload,
    inference_row_create_many_multi_mixed_actuals_payload,
    inference_row_create_many_reg_payload,
)
import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("inference_rows_create"))
def test_inference_row_create(client, api_key):
    response = client.post(
        "/v1/inference-rows",
        json={
            **inference_row_create_single_row_payload,
            "model_id": state.model_multi["id"],
        },
        headers={"api-key": api_key},
    )
    state.inference_row_multi = response.json()
    assert response.status_code == status.HTTP_201_CREATED
    validated = schemas.InferenceRow(**response.json())


@pytest.mark.order(get_order_number("inference_rows_create_many"))
def test_inference_row_create_many(client, api_key):
    response_binary = client.post(
        "/v1/inference-rows/batch",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_binary["id"]},
                inference_row_create_many_binary_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    state.inference_row_binary = response_binary.json()[0]

    response_multi = client.post(
        "/v1/inference-rows/batch",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi["id"]},
                inference_row_create_many_multi_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_multi_2 = client.post(
        "/v1/inference-rows/batch",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi_2["id"]},
                inference_row_create_many_multi_no_actual_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_multi_3 = client.post(
        "/v1/inference-rows/batch",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_multi_3["id"]},
                inference_row_create_many_multi_mixed_actuals_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    response_reg = client.post(
        "/v1/inference-rows/batch",
        json=list(
            map(
                lambda x: {**x, "model_id": state.model_regression["id"]},
                inference_row_create_many_reg_payload,
            )
        ),
        headers={"api-key": api_key},
    )

    assert response_binary.status_code == status.HTTP_201_CREATED
    assert response_multi.status_code == status.HTTP_201_CREATED
    assert response_reg.status_code == status.HTTP_201_CREATED
    validated = [schemas.InferenceRow(**m) for m in response_binary.json()]
    validated = [schemas.InferenceRow(**m) for m in response_multi.json()]
    validated = [schemas.InferenceRow(**m) for m in response_multi_2.json()]
    validated = [schemas.InferenceRow(**m) for m in response_multi_3.json()]
    validated = [schemas.InferenceRow(**m) for m in response_reg.json()]


@pytest.mark.order(get_order_number("inference_rows_get_model's_all"))
def test_inference_row_get_models_all(client, api_key):
    response_multi = client.get(
        f"/v1/inference-rows?model_id={state.model_multi['id']}",
        headers={"api-key": api_key},
    )
    response_binary = client.get(
        f"/v1/inference-rows?model_id={state.model_binary['id']}",
        headers={"api-key": api_key},
    )
    response_wrong_model = client.get(
        f"/v1/inference-rows?model_id=wrong_model_id",
        headers={"api-key": api_key},
    )
    assert response_multi.status_code == status.HTTP_200_OK
    assert response_binary.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND
    validated = [schemas.InferenceRow(**m) for m in response_multi.json()]
    validated = [schemas.InferenceRow(**m) for m in response_binary.json()]


@pytest.mark.order(get_order_number("inference_rows_get"))
def test_inference_row_get(client, api_key):
    response = client.get(
        f"/v1/inference-rows/{state.inference_row_multi['id']}",
        headers={"api-key": api_key},
    )
    response_wrong_inference = client.get(
        f"/v1/inference-rows/wrong_inference_id",
        headers={"api-key": api_key},
    )
    assert response.status_code == status.HTTP_200_OK
    assert response_wrong_inference.status_code == status.HTTP_404_NOT_FOUND
    validated = schemas.InferenceRow(**response.json())


@pytest.mark.order(get_order_number("inference_rows_xai"))
def test_inference_row_xai(client, api_key):
    response = client.get(
        f"/v1/inference-rows/{state.inference_row_binary['id']}/xai",
        headers={"api-key": api_key},
    )

    response_wrong_inference = client.get(
        f"/v1/inference-rows/wrong_inference_id/xai",
        headers={"api-key": api_key},
    )

    assert response.status_code == status.HTTP_200_OK
    assert response_wrong_inference.status_code == status.HTTP_404_NOT_FOUND



from typing import Any, List
from sqlalchemy.orm import Session
from whitebox.crud.base import CRUDBase
from whitebox.entities.ModelIntegrityMetric import (
    ModelIntegrityMetric as ModelIntegrityMetricEntity,
)
from whitebox.schemas.modelIntegrityMetric import (
    ModelIntegrityMetricCreate,
    ModelIntegrityMetric,
)


class CRUD(CRUDBase[ModelIntegrityMetric, ModelIntegrityMetricCreate, Any]):
    def get_model_integrity_metrics_by_model(
        self, db: Session, *, model_id: str
    ) -> List[ModelIntegrityMetric]:
        return (
            db.query(self.model)
            .filter(ModelIntegrityMetricEntity.model_id == model_id)
            .all()
        )


model_integrity_metrics = CRUD(ModelIntegrityMetricEntity)



from typing import List, Union
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.performanceMetric import (
    BinaryClassificationMetrics,
    MultiClassificationMetrics,
    RegressionMetrics,
)
from whitebox.schemas.user import User
from whitebox.schemas.model import ModelType
from whitebox.utils.errors import add_error_responses, errors


performance_metrics_router = APIRouter()


@performance_metrics_router.get(
    "/performance-metrics",
    tags=["Performance Metrics"],
    response_model=Union[
        List[BinaryClassificationMetrics],
        List[MultiClassificationMetrics],
        List[RegressionMetrics],
    ],
    summary="Get all model's performance metrics",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_all_models_performance_metrics(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches the performance metrics of a specific model. A model id is required."""

    model = crud.models.get(db, model_id)
    if model:
        if vars(model)["type"] == ModelType.binary:
            return crud.binary_classification_metrics.get_by_model(
                db=db, model_id=model_id
            )
        elif vars(model)["type"] == ModelType.multi_class:
            return crud.multi_classification_metrics.get_by_model(
                db=db, model_id=model_id
            )
        elif vars(model)["type"] == ModelType.regression:
            return crud.regression_metrics.get_by_model(db=db, model_id=model_id)
    else:
        return errors.not_found("Model not found")



from .Alert import Alert
from .DatasetRow import DatasetRow
from .DriftingMetric import DriftingMetric
from .Inference import InferenceRow
from .Model import Model
from .ModelIntegrityMetric import ModelIntegrityMetric
from .ModelMonitor import ModelMonitor
from .PerformanceMetric import (
    BinaryClassificationMetrics,
    MultiClassificationMetrics,
    RegressionMetrics,
)
from .User import User



from fastapi import FastAPI, Depends
import asyncio
import json
from whitebox.utils.logger import cronLogger as logger

from whitebox.core.settings import get_cron_settings
from whitebox.cron_tasks.tasks import task_manager
from fastapi.openapi.utils import get_openapi


settings = get_cron_settings()
cron_app = FastAPI(title=settings.APP_NAME_CRON, redoc_url="/")


@cron_app.on_event("startup")
async def init():
    # Start task amanager
    asyncio.get_event_loop().create_task(task_manager.run())


@cron_app.on_event("shutdown")
async def shutdown():
    logger.info("App is shutting down...")
    logger.info("Task Manager is shutting down...")
    await task_manager.shutdown()


def app_openapi():
    if cron_app.openapi_schema:
        return cron_app.openapi_schema
    openapi_schema = get_openapi(
        title="Cron API", version=settings.VERSION, routes=cron_app.routes
    )

    cron_app.openapi_schema = openapi_schema
    return cron_app.openapi_schema


cron_app.openapi = app_openapi



from enum import Enum
import numpy as np
import pandas as pd
from whitebox.schemas.model import ModelCreateDto, ModelType
from typing import Dict, Optional
import requests
import logging
from fastapi import status

from whitebox.schemas.modelMonitor import (
    AlertSeverity,
    ModelMonitorCreateDto,
    MonitorMetrics,
    MonitorStatus,
)


class APiVersion(str, Enum):
    v1 = "v1"


logger = logging.getLogger(__name__)


class Whitebox:
    def __init__(
        self,
        host: str,
        api_key: str,
        verbose: bool = False,
        api_version: APiVersion = APiVersion.v1,
    ):
        self.host = host
        self.api_key = api_key
        self.verbose = verbose
        self.api_version = api_version

    def create_model(
        self,
        name: str,
        type: ModelType,
        prediction: str,
        labels: Dict[str, int] = None,
        description: str = "",
    ) -> dict:
        """
        Create a new model in Whitebox and define a type, and a schema for it.
        """
        new_model = ModelCreateDto(
            name=name,
            description=description,
            type=type,
            labels=labels,
            prediction=prediction,
        )
        result = requests.post(
            url=f"{self.host}/{self.api_version}/models",
            json=new_model.dict(),
            headers={"api-key": self.api_key},
        )

        logger.info(result.json())
        return result.json()

    def get_model(self, model_id: str) -> dict:
        """
        Returns a model by its id. If the model does not exist, returns None.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/models/{model_id}",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def delete_model(self, model_id: str):
        """
        Deletes a model by its id. If any error occurs, returns False.
        """
        result = requests.delete(
            url=f"{self.host}/{self.api_version}/models/{model_id}",
            headers={"api-key": self.api_key},
        )

        if result.status_code == status.HTTP_200_OK:
            return True

        return False

    def log_training_dataset(
        self, model_id: str, non_processed: pd.DataFrame, processed: pd.DataFrame
    ) -> bool:
        """
        Logs a training dataset for a model.

        Non processed is a dataframe with the raw data.
        Processed is a dataframe with the data after it has been processed and before it has entered the model.
        """
        self._check_processed_and_non_processed_length(processed, non_processed)
        non_processed_json = non_processed.to_dict(orient="records")
        processed_json = processed.to_dict(orient="records")

        dataset_rows = []
        for i in range(len(non_processed)):
            dataset_rows.append(
                dict(
                    model_id=model_id,
                    nonprocessed=non_processed_json[i],
                    processed=processed_json[i],
                )
            )

        result = requests.post(
            url=f"{self.host}/{self.api_version}/dataset-rows",
            headers={"api-key": self.api_key},
            json=dataset_rows,
        )
        if result.status_code == status.HTTP_201_CREATED:
            return True

        return False

    def log_inferences(
        self,
        model_id: str,
        non_processed: pd.DataFrame,
        processed: pd.DataFrame,
        timestamps: pd.Series,
        actuals: pd.Series = None,
    ) -> bool:
        """
        Logs inferences of a model.

        Non processed is a dataframe with the raw data.
        Processed is a dataframe with the data after it has been processed and before it has entered the model.
        """
        self._check_processed_and_non_processed_length(processed, non_processed)
        non_processed_json = non_processed.to_dict(orient="records")
        processed_json = processed.to_dict(orient="records")
        timestamps_list = timestamps.tolist()

        if actuals is not None:
            actuals = actuals.replace({np.nan: None})
            actuals_list = actuals.tolist()
        else:
            actuals_list = actuals

        inference_rows = []
        for i in range(len(non_processed)):
            inference_rows.append(
                dict(
                    model_id=model_id,
                    nonprocessed=non_processed_json[i],
                    processed=processed_json[i],
                    timestamp=timestamps_list[i],
                    actual=actuals_list[i] if actuals_list is not None else None,
                )
            )

        result = requests.post(
            url=f"{self.host}/{self.api_version}/inference-rows/batch",
            headers={"api-key": self.api_key},
            json=inference_rows,
        )
        if result.status_code == status.HTTP_201_CREATED:
            return True

        return False

    def get_xai_row(self, inference_row_id: str):
        """
        Given a specific inference row id, this endpoint produces an explainability report for this inference.
        If some of the required data isn't found, returns None.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/inference-rows/{inference_row_id}/xai",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def create_model_monitor(
        self,
        model_id: str,
        name: str,
        status: MonitorStatus,
        metric: MonitorMetrics,
        severity: AlertSeverity,
        email: str,
        feature: Optional[str],
        lower_threshold: Optional[float],
    ) -> dict:
        """
        Creates a monitor for a model.
        """

        model_monitor = ModelMonitorCreateDto(
            model_id=model_id,
            name=name,
            status=status,
            metric=metric,
            feature=feature,
            lower_threshold=lower_threshold,
            severity=severity,
            email=email,
        )

        result = requests.post(
            url=f"{self.host}/{self.api_version}/model-monitors",
            json=model_monitor.dict(),
            headers={"api-key": self.api_key},
        )

        logger.info(result.json())
        return result.json()

    def get_alerts(self, model_id: str = "") -> dict:
        """
        Returns all alerts for a model.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/alerts?modelId={model_id}",
            headers={"api-key": self.api_key},
        )

        logger.info(result.json())
        return result.json()

    def get_drifting_metrics(self, model_id: str) -> dict:
        """
        Returns a model's drifting metric reports. If the model does not exist, returns None.
        If the model exists but there are no metrics, returns an empty list.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/drifting-metrics?model_id={model_id}",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def get_descriptive_statistics(self, model_id: str) -> dict:
        """
        Returns a model's descriptive statistics reports. If the model does not exist, returns None.
        If the model exists but there are no metrics, returns an empty list.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/model-integrity-metrics?model_id={model_id}",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def get_performance_metrics(self, model_id: str) -> dict:
        """
        Returns a model's performance metric reports. If the model does not exist, returns None.
        If the model exists but there are no metrics, returns an empty list.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/performance-metrics?model_id={model_id}",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def _check_processed_and_non_processed_length(
        self, processed: pd.DataFrame, non_processed: pd.DataFrame
    ) -> bool:
        """
        Checks if the processed and non processed dataframes have the same number of rows.
        """
        if len(processed) != len(non_processed):
            raise ValueError(
                "Processed and non processed dataframes must have the same length."
            )
        return True



from whitebox.tests.v1.mock_data import (
    model_binary_create_payload,
    model_multi_create_payload,
    model_multi_2_create_payload,
    model_multi_3_create_payload,
    model_regression_create_payload,
    model_update_payload,
)
import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("models_create"))
def test_model_create(client, api_key):
    response_binary = client.post(
        "/v1/models",
        json={**model_binary_create_payload},
        headers={"api-key": api_key},
    )

    response_multi = client.post(
        "/v1/models",
        json={**model_multi_create_payload},
        headers={"api-key": api_key},
    )

    response_multi_2 = client.post(
        "/v1/models",
        json={**model_multi_2_create_payload},
        headers={"api-key": api_key},
    )

    response_multi_3 = client.post(
        "/v1/models",
        json={**model_multi_3_create_payload},
        headers={"api-key": api_key},
    )

    response_regression = client.post(
        "/v1/models",
        json={**model_regression_create_payload},
        headers={"api-key": api_key},
    )

    state.model_binary = response_binary.json()
    state.model_multi = response_multi.json()
    state.model_multi_2 = response_multi_2.json()
    state.model_multi_3 = response_multi_3.json()
    state.model_regression = response_regression.json()

    assert response_binary.status_code == status.HTTP_201_CREATED
    assert response_multi.status_code == status.HTTP_201_CREATED
    assert response_multi_2.status_code == status.HTTP_201_CREATED
    assert response_multi_3.status_code == status.HTTP_201_CREATED
    assert response_regression.status_code == status.HTTP_201_CREATED

    validated = schemas.Model(**response_binary.json())
    validated = schemas.Model(**response_multi.json())
    validated = schemas.Model(**response_multi_2.json())
    validated = schemas.Model(**response_multi_3.json())
    validated = schemas.Model(**response_regression.json())


@pytest.mark.order(get_order_number("models_get_all"))
def test_model_get_all(client, api_key):
    response = client.get(f"/v1/models", headers={"api-key": api_key})
    assert response.status_code == status.HTTP_200_OK
    validated = [schemas.Model(**m) for m in response.json()]


@pytest.mark.order(get_order_number("models_get"))
def test_model_get(client, api_key):
    response = client.get(
        f"/v1/models/{state.model_multi['id']}", headers={"api-key": api_key}
    )
    response_wrong_model = client.get(
        f"/v1/models/wrong_model_id", headers={"api-key": api_key}
    )

    assert response.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = schemas.Model(**response.json())


@pytest.mark.order(get_order_number("models_update"))
def test_model_update(client, api_key):
    response = client.put(
        f"/v1/models/{state.model_multi['id']}",
        json=model_update_payload,
        headers={"api-key": api_key},
    )
    response_wrong_model = client.put(
        f"/v1/models/wrong_model_id",
        json=model_update_payload,
        headers={"api-key": api_key},
    )

    assert response.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = schemas.Model(**response.json())


@pytest.mark.order(get_order_number("models_delete"))
def test_model_delete(client, api_key):
    response_binary = client.delete(
        f"/v1/models/{state.model_binary['id']}", headers={"api-key": api_key}
    )
    response_multi = client.delete(
        f"/v1/models/{state.model_multi['id']}", headers={"api-key": api_key}
    )
    response_multi_2 = client.delete(
        f"/v1/models/{state.model_multi_2['id']}", headers={"api-key": api_key}
    )
    response_multi_3 = client.delete(
        f"/v1/models/{state.model_multi_3['id']}", headers={"api-key": api_key}
    )
    response_regression = client.delete(
        f"/v1/models/{state.model_regression['id']}", headers={"api-key": api_key}
    )
    response_no_model = client.delete(
        f"/v1/models/{state.model_binary['id']}", headers={"api-key": api_key}
    )

    assert response_binary.status_code == status.HTTP_200_OK
    assert response_multi.status_code == status.HTTP_200_OK
    assert response_multi_2.status_code == status.HTTP_200_OK
    assert response_multi_3.status_code == status.HTTP_200_OK
    assert response_regression.status_code == status.HTTP_200_OK
    assert response_no_model.status_code == status.HTTP_404_NOT_FOUND



import pytest
from whitebox.tests.v1.mock_data import (
    model_monitor_accuracy_create_payload,
    model_monitor_f1_create_payload,
    model_monitor_data_drift_create_payload,
    model_monitor_concept_drift_create_payload,
    model_monitor_precision_create_payload,
    model_monitor_r_square_create_payload,
)
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("model_monitor_create"))
def test_model_monitor_create(client, api_key):
    accuracy_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_accuracy_create_payload,
            "model_id": state.model_multi["id"],
        },
        headers={"api-key": api_key},
    )

    f1_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_f1_create_payload,
            "model_id": state.model_multi["id"],
        },
        headers={"api-key": api_key},
    )

    data_drift_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_data_drift_create_payload,
            "model_id": state.model_binary["id"],
        },
        headers={"api-key": api_key},
    )

    concept_drift_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_concept_drift_create_payload,
            "model_id": state.model_binary["id"],
        },
        headers={"api-key": api_key},
    )

    precision_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_precision_create_payload,
            "model_id": state.model_binary["id"],
        },
        headers={"api-key": api_key},
    )

    r_square_monitor = client.post(
        "/v1/model-monitors",
        json={
            **model_monitor_r_square_create_payload,
            "model_id": state.model_regression["id"],
        },
        headers={"api-key": api_key},
    )

    assert accuracy_monitor.status_code == status.HTTP_201_CREATED
    validated = schemas.ModelMonitor(**accuracy_monitor.json())


@pytest.mark.order(get_order_number("model_monitors_get_model_all"))
def test_model_monitors_get_model_all(client, api_key):
    response_multi = client.get(
        f"/v1/model-monitors?model_id={state.model_multi['id']}",
        headers={"api-key": api_key},
    )
    response_all = client.get(
        f"/v1/model-monitors",
        headers={"api-key": api_key},
    )
    response_wrong_model = client.get(
        f"/v1/model-monitors?model_id=wrong_model_id",
        headers={"api-key": api_key},
    )

    assert len(response_multi.json()) == 2
    assert len(response_all.json()) == 6

    assert response_multi.status_code == status.HTTP_200_OK
    assert response_all.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = [schemas.ModelMonitor(**m) for m in response_multi.json()]
    validated = [schemas.ModelMonitor(**m) for m in response_all.json()]



from typing import List
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.model import Model, ModelCreateDto, ModelUpdateDto
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.schemas.utils import StatusCode
from whitebox.schemas.user import User
from whitebox.utils.errors import add_error_responses, errors


models_router = APIRouter()


@models_router.post(
    "/models",
    tags=["Models"],
    response_model=Model,
    summary="Create model",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 401]),
)
async def create_model(
    body: ModelCreateDto,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> Model:
    """Inserts a model into the database"""

    new_model = crud.models.create(db=db, obj_in=body)
    return new_model


@models_router.get(
    "/models",
    tags=["Models"],
    response_model=List[Model],
    summary="Get all models",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401]),
)
async def get_all_models(
    db: Session = Depends(get_db), authenticated_user: User = Depends(authenticate_user)
):
    """Fetches all models from the database"""

    models_in_db = crud.models.get_all(db=db)
    return models_in_db


@models_router.get(
    "/models/{model_id}",
    tags=["Models"],
    response_model=Model,
    summary="Get model by id",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_model(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches the model with the specified id from the database"""

    model = crud.models.get(db=db, _id=model_id)

    if not model:
        return errors.not_found("Model not found")

    return model


@models_router.put(
    "/models/{model_id}",
    tags=["Models"],
    response_model=Model,
    summary="Update model",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([400, 401, 404]),
)
async def update_model(
    model_id: str,
    body: ModelUpdateDto,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> Model:
    """Updates record of the model with the specified id"""

    model = crud.models.get(db=db, _id=model_id)

    if not model:
        return errors.not_found("Model not found")

    return crud.models.update(db=db, db_obj=model, obj_in=body)


@models_router.delete(
    "/models/{model_id}",
    tags=["Models"],
    response_model=StatusCode,
    summary="Delete user",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def delete_model(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> StatusCode:
    """Deletes the model with the specified id from the database"""

    model = crud.models.get(db=db, _id=model_id)
    if not model:
        return errors.not_found("Model not found")

    crud.models.remove(db=db, _id=model_id)
    return {"status_code": status.HTTP_200_OK}



import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("drifting_metrics_get_model_all"))
def test_drifting_metric_get_model_all(client, api_key):
    response_multi = client.get(
        f"/v1/drifting-metrics?model_id={state.model_multi['id']}",
        headers={"api-key": api_key},
    )

    response_binary = client.get(
        f"/v1/drifting-metrics?model_id={state.model_binary['id']}",
        headers={"api-key": api_key},
    )
    response_wrong_model = client.get(
        f"/v1/drifting-metrics?model_id=wrong_model_id",
        headers={"api-key": api_key},
    )

    response_binary_json = response_binary.json()
    response_multi_json = response_multi.json()

    assert len(response_multi_json) == 1
    assert len(response_binary_json) == 1

    assert response_multi.status_code == status.HTTP_200_OK
    assert response_binary.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = [schemas.DriftingMetric(**m) for m in response_multi_json]
    validated = [schemas.DriftingMetric(**m) for m in response_binary_json]



from whitebox.schemas.utils import HealthCheck
from fastapi import APIRouter, status
from whitebox.cron_tasks.monitoring_metrics import run_calculate_metrics_pipeline
from whitebox.cron_tasks.monitoring_alerts import run_create_alerts_pipeline


cron_tasks_router = APIRouter()


@cron_tasks_router.post(
    "/cron-tasks/run",
    tags=["Cron Tasks"],
    summary="Helper endpoint",
    status_code=status.HTTP_200_OK,
    response_description="Result of cron tasks",
)
async def run_cron():
    """A helper endpoint that triggers the metrics and alerts pipelines while testing."""

    await run_calculate_metrics_pipeline()
    await run_create_alerts_pipeline()
    return HealthCheck(status="OK")



from typing import Dict, Optional
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase
import enum


class ModelType(str, enum.Enum):
    binary = "binary"
    multi_class = "multi_class"
    regression = "regression"


class ModelBase(BaseModel):
    name: str
    description: str
    type: ModelType
    prediction: str
    labels: Optional[Dict[str, int]]


class Model(ModelBase, ItemBase):
    pass


class ModelCreateDto(ModelBase):
    pass


class ModelUpdateDto(BaseModel):
    name: Optional[str]
    description: Optional[str]
    type: Optional[ModelType]



from whitebox.schemas.utils import ErrorResponse


tags_metadata = [
    {
        "name": "Health",
        "description": "Health endpoints are used for checking the status of the service",
    },
    {
        "name": "Models",
        "description": "This set of endpoints handles the models that a user creates.",
    },
    {
        "name": "Dataset Rows",
        "description": "This set of endpoints handles the dataset rows.",
    },
    {
        "name": "Inference Rows",
        "description": "This set of endpoints handles a model's inference rows.",
    },
    {
        "name": "Performance Metrics",
        "description": "This set of endpoints handles a model's performance metrics.",
    },
    {
        "name": "Drifting Metrics",
        "description": "This set of endpoints handles a model's drifting metrics.",
    },
    {
        "name": "Model Integrity Metrics",
        "description": "This set of endpoints handles a model's integrity metrics.",
    },
    {
        "name": "Model Monitors",
        "description": "This set of endpoints handles a model's model monitors.",
    },
    {
        "name": "Alerts",
        "description": "This set of endpoints handles a model's alerts.",
    },
    {
        "name": "Users",
        "description": "This is a helper endpoint to create and delete an admin user in tests.",
    },
    {
        "name": "Cron Tasks",
        "description": "This is a helper endpoint to trigger cron tasks for tests.",
    },
]


bad_request: ErrorResponse = {
    "title": "BadRequest",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

validation_error: ErrorResponse = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

authorization_error: ErrorResponse = {
    "title": "AuthorizationError",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

not_found_error: ErrorResponse = {
    "title": "NotFoundError",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

conflict_error: ErrorResponse = {
    "title": "ConflictError",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

conflict_error: ErrorResponse = {
    "title": "ConflictError",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}

content_gone: ErrorResponse = {
    "title": "ContentGone",
    "type": "object",
    "properties": {
        "error": {"title": "Error Message", "type": "string"},
        "status_code": {"title": "Status code", "type": "integer"},
    },
}



from setuptools import setup, find_packages


VERSION = "0.0.19"

DESCRIPTION = "Whitebox is an open source E2E ML monitoring platform with edge capabilities that plays nicely with kubernetes"
LONG_DESCRIPTION = ""

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="whitebox-sdk",
    version=VERSION,
    author="Squaredev",
    author_email="hello@squaredev.io",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'
    keywords=["python", "model monitoring", "whitebox", "mlops"],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)



from datetime import datetime
from typing import Any, Dict, Union, Optional, overload
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase


class InferenceRowBase(BaseModel):
    model_id: str
    timestamp: Union[str, datetime]
    # Prediction is included into nonprocessed & processed
    nonprocessed: Dict[str, Any]
    processed: Dict[str, float]
    actual: Optional[float]


class InferenceRow(InferenceRowBase, ItemBase):
    pass


class InferenceRowCreateDto(InferenceRowBase):
    pass



from typing import Any, Dict
from sqlalchemy.ext.declarative import as_declarative, declared_attr


class_registry: Dict = {}


@as_declarative(class_registry=class_registry)
class Base:
    id: Any
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()



from sqlalchemy import Column, Integer, Float, ForeignKey, String, DateTime, JSON
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid


class BinaryClassificationMetrics(Base):
    __tablename__ = "binary_classification_metrics"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    accuracy = Column(Float)
    precision = Column(Float)
    recall = Column(Float)
    f1 = Column(Float)
    true_negative = Column(Integer)
    false_positive = Column(Integer)
    false_negative = Column(Integer)
    true_positive = Column(Integer)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)


class MultiClassificationMetrics(Base):
    __tablename__ = "multi_classification_metrics"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    accuracy = Column(Float)
    precision = Column(JSON)
    recall = Column(JSON)
    f1 = Column(JSON)
    confusion_matrix = Column(JSON)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)


class RegressionMetrics(Base):
    __tablename__ = "regression_metrics"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    r_square = Column(Float)
    mean_squared_error = Column(Float)
    mean_absolute_error = Column(Float)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)



from datetime import datetime
from typing import Dict, List, Union
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase


class ColumnDataDriftMetrics(BaseModel):
    """One column drift metrics"""

    column_name: str
    column_type: str
    stattest_name: str
    drift_score: float
    drift_detected: bool
    threshold: float


class DataDriftTable(BaseModel):
    number_of_columns: int
    number_of_drifted_columns: int
    share_of_drifted_columns: float
    dataset_drift: bool
    drift_by_columns: Dict[str, ColumnDataDriftMetrics]


class CramerV(BaseModel):
    """CramerV statistics"""

    column_name: str
    kind: str
    values: Dict[str, List[str]]


class ColumnConceptDriftCorrelationMetrics(BaseModel):
    """One column concept drift correlation metrics"""

    column_name: str
    current: Dict[str, CramerV]
    reference: Dict[str, CramerV]


class ColumnConceptDriftMetrics(BaseModel):
    """One column concept drift metrics"""

    column_name: str
    column_type: str
    stattest_name: str
    drift_score: float
    drift_detected: bool
    stattest_threshold: float


class ConceptDriftTable(BaseModel):
    """Concept drift Table metrics"""

    concept_drift_summary: ColumnConceptDriftMetrics
    column_correlation: ColumnConceptDriftCorrelationMetrics


class DriftingMetricBase(ItemBase):
    model_id: str
    timestamp: Union[str, datetime]
    concept_drift_summary: ConceptDriftTable
    data_drift_summary: DataDriftTable


class DriftingMetric(DriftingMetricBase, ItemBase):
    pass



from typing import List
from fastapi import status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException as StarletteHTTPException
from fastapi.exceptions import HTTPException
from starlette.requests import Request
from whitebox.schemas.utils import ErrorProps
from whitebox.utils.logger import log


class CustomError(BaseException):
    async def http_exception_handler(
        self, _: Request, exc: StarletteHTTPException
    ) -> ErrorProps:
        log.error(f"{exc.status_code}: {exc.detail}")
        return JSONResponse(
            status_code=exc.status_code,
            content=jsonable_encoder(
                {"error": f"{str(exc.detail)}", "status_code": exc.status_code}
            ),
        )

    async def validation_exception_handler(
        self,
        _: Request,
        exc: HTTPException,
    ) -> ErrorProps:
        responsible_value = exc.errors()[0]["loc"][-1]
        reason = exc.errors()[0]["msg"]
        log.error(
            f"{status.HTTP_422_UNPROCESSABLE_ENTITY}: ({str(responsible_value)}) {str(reason)}"
        )
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=jsonable_encoder(
                {
                    "error": f"({str(responsible_value)}) {str(reason)}",
                    "status_code": status.HTTP_422_UNPROCESSABLE_ENTITY,
                }
            ),
        )

    def bad_request(self, msg: str = "Bad request") -> ErrorProps:
        log.error(f"{status.HTTP_400_BAD_REQUEST}: {str(msg)}")
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content=jsonable_encoder(
                {"error": str(msg), "status_code": status.HTTP_400_BAD_REQUEST}
            ),
        )

    def not_found(self, msg: str = "Content not found") -> ErrorProps:
        log.error(f"{status.HTTP_404_NOT_FOUND}: {str(msg)}")
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content=jsonable_encoder(
                {"error": str(msg), "status_code": status.HTTP_404_NOT_FOUND}
            ),
        )


errors = CustomError()


def add_error_responses(status_codes) -> List[ErrorProps]:
    """
    For the schema to work the part after schemas/ should correspond to a title error schema in whitebox/api/app/v1/docs.py
    """

    error_responses = {
        400: {
            "description": "Bad Request",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/BadRequest"}
                }
            },
        },
        401: {
            "description": "Authorization Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/AuthorizationError"}
                }
            },
        },
        404: {
            "description": "Not Found Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/NotFoundError"}
                }
            },
        },
        409: {
            "description": "Conflict Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/ConflictError"}
                }
            },
        },
        410: {
            "description": "Content Gone",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/ContenGone"}
                }
            },
        },
        422: {
            "description": "Validation Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/HTTPValidationError"}
                }
            },
        },
    }
    responses = {}
    for code in status_codes:
        responses[code] = error_responses[code]
    return responses



import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("alerts_get"))
def test_alerts_get_model_all(client, api_key):
    response_all = client.get(
        f"/v1/alerts",
        headers={"api-key": api_key},
    )

    response_model_all = client.get(
        f"/v1/alerts?model_id={state.model_multi['id']}",
        headers={"api-key": api_key},
    )

    response_wrong_model = client.get(
        f"/v1/alerts?model_id=wrong_model_id",
        headers={"api-key": api_key},
    )

    assert len(response_all.json()) == 6

    assert response_all.status_code == status.HTTP_200_OK
    assert response_model_all.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = [schemas.Alert(**m) for m in response_all.json()]
    validated = [schemas.Alert(**m) for m in response_model_all.json()]



from typing import Dict, List
from whitebox.middleware.auth import authenticate_user
from whitebox.schemas.inferenceRow import InferenceRow, InferenceRowCreateDto
from whitebox.analytics.xai_models.pipelines import (
    create_xai_pipeline_per_inference_row,
)
import pandas as pd
from whitebox.schemas.user import User
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.utils.errors import add_error_responses, errors


inference_rows_router = APIRouter()


@inference_rows_router.post(
    "/inference-rows",
    tags=["Inference Rows"],
    response_model=InferenceRow,
    summary="Create an inference row",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 401]),
)
async def create_row(
    body: InferenceRowCreateDto,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> InferenceRow:
    """Inserts an inference row into the database."""

    new_inference_row = crud.inference_rows.create(db=db, obj_in=body)
    return new_inference_row


@inference_rows_router.post(
    "/inference-rows/batch",
    tags=["Inference Rows"],
    response_model=List[InferenceRow],
    summary="Create many inference rows",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 401]),
)
async def create_many_inference_rows(
    body: List[InferenceRowCreateDto],
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
) -> List[InferenceRow]:
    """Inserts a set of inference rows into the database."""

    new_inference_rows = crud.inference_rows.create_many(db=db, obj_list=body)
    return new_inference_rows


@inference_rows_router.get(
    "/inference-rows",
    tags=["Inference Rows"],
    response_model=List[InferenceRow],
    summary="Get all model's inference rows",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_all_models_inference_rows(
    model_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches the inference rows of a specific model. A model id is required."""

    model = crud.models.get(db, model_id)
    if model:
        return crud.inference_rows.get_inference_rows_by_model(db=db, model_id=model_id)
    else:
        return errors.not_found("Model not found")


@inference_rows_router.get(
    "/inference-rows/{inference_row_id}",
    tags=["Inference Rows"],
    response_model=InferenceRow,
    summary="Get inference row by id",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def get_inference_row(
    inference_row_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """Fetches a specific inference row. An inference row id is required."""

    inference_row = crud.inference_rows.get(db=db, _id=inference_row_id)
    if not inference_row:
        return errors.not_found("Inference not found")

    return inference_row


@inference_rows_router.get(
    "/inference-rows/{inference_row_id}/xai",
    tags=["Inference Rows"],
    response_model=Dict[str, float],
    summary="Creates and fetches am explainability report for an inference row",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([401, 404]),
)
async def create_inference_row_xai_report(
    inference_row_id: str,
    db: Session = Depends(get_db),
    authenticated_user: User = Depends(authenticate_user),
):
    """
    Given a specific inference row id, this endpoint produces an explainability report for this inference.
    The XAI pipeline requires a set of dataset rows as a training set, a model and the inference row.
    If one of those three is not found in the database, a 404 error is returned.
    """

    inference_row = crud.inference_rows.get(db=db, _id=inference_row_id)
    if not inference_row:
        return errors.not_found(f"Inference row with id {inference_row_id} not found")

    inference_row_df = pd.DataFrame([inference_row.processed])
    inference_row_series = inference_row_df.drop(columns=["target"]).iloc[0]

    model = crud.models.get(db=db, _id=inference_row.model_id)
    if not model:
        return errors.not_found(f"Model with id {inference_row.model_id} not found")

    dataset_rows = crud.dataset_rows.get_dataset_rows_by_model(db=db, model_id=model.id)
    if not dataset_rows:
        return errors.not_found(
            f"Dataset rows for model with id {inference_row.model_id} not found"
        )

    dataset_rows_processed = [x.processed for x in dataset_rows]

    xai_report = create_xai_pipeline_per_inference_row(
        training_set=pd.DataFrame(dataset_rows_processed),
        target=model.prediction,
        inference_row=inference_row_series,
        type_of_task=model.type,
        model_id=model.id,
    )

    return xai_report



from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from sqlalchemy.orm import Session
from whitebox.core.db import Base
import datetime

ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        """
        self.model = model

    def get(self, db: Session, _id: str) -> Optional[ModelType]:
        return db.query(self.model).filter(self.model.id == _id).first()

    def get_all(
        self, db: Session, *, skip: int = 0, limit: int = 100
    ) -> List[ModelType]:
        return db.query(self.model).offset(skip).limit(limit).all()

    def get_first_by_filter(
        self, db: Session, **kwargs: Union[Dict[str, str], str]
    ) -> ModelType:
        return db.query(self.model).filter_by(**kwargs).first()

    def create(self, db: Session, *, obj_in: CreateSchemaType) -> ModelType:
        date_now = datetime.datetime.utcnow()
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(**obj_in_data, created_at=date_now, updated_at=date_now)
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def create_many(
        self, db: Session, *, obj_list: List[CreateSchemaType]
    ) -> List[ModelType]:
        date_now = datetime.datetime.utcnow()
        obj_list_in_data = jsonable_encoder(obj_list)
        db_obj_list = list(
            map(
                lambda x: self.model(**x, created_at=date_now, updated_at=date_now),
                obj_list_in_data,
            )
        )
        db.add_all(db_obj_list)
        db.commit()
        for obj in db_obj_list:
            db.refresh(obj)
        return db_obj_list

    def update(
        self,
        db: Session,
        *,
        db_obj: ModelType,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]]
    ) -> ModelType:
        date_now = datetime.datetime.utcnow()
        obj_data = jsonable_encoder(db_obj)
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.dict(exclude_unset=True)
        for field in obj_data:
            if field in update_data:
                setattr(db_obj, field, update_data[field])
        setattr(db_obj, "updated_at", date_now)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def remove(self, db: Session, *, _id: str):
        db.query(self.model).filter(self.model.id == _id).delete()
        db.commit()
        return



from typing import Any, List
from sqlalchemy.orm import Session
from whitebox.crud.base import CRUDBase
from whitebox.schemas.inferenceRow import InferenceRow, InferenceRowCreateDto
from whitebox.entities.Inference import InferenceRow as InferenceRowEntity


class CRUD(CRUDBase[InferenceRow, InferenceRowCreateDto, Any]):
    def get_inference_rows_by_model(
        self, db: Session, *, model_id: str
    ) -> List[InferenceRow]:
        return (
            db.query(self.model).filter(InferenceRowEntity.model_id == model_id).all()
        )


inference_rows = CRUD(InferenceRowEntity)



from sqlalchemy import Column, String, DateTime, JSON, Enum
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid
from sqlalchemy.orm import relationship
from whitebox.schemas.model import ModelType


class Model(Base):
    __tablename__ = "models"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    name = Column(String)
    description = Column(String)
    type = Column("type", Enum(ModelType))
    labels = Column(JSON, nullable=True)
    prediction = Column(String)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)

    dataset_rows = relationship("DatasetRow")
    inference_rows = relationship("InferenceRow")
    binary_classification_metrics = relationship("BinaryClassificationMetrics")
    multi_classification_metrics = relationship("MultiClassificationMetrics")
    regression_metrics = relationship("RegressionMetrics")
    drifting_metrics = relationship("DriftingMetric")
    model_integrity_metrics = relationship("ModelIntegrityMetric")
    model_monitors = relationship("ModelMonitor")
